# Written by Lauri Lehtovaara, 2007

"""This module implements time propagators for time-dependent density
functional theory calculations."""

import numpy as np

from gpaw.utilities.blas import axpy

from gpaw.tddft.utils import MultiBlas
from gpaw.tddft.tdopers import DummyDensity

###############################################################################
# DummyKPoint
###############################################################################
class DummyKPoint(object):
    __slots__ = ('psit_nG',)

###############################################################################
# DummyPropagator
###############################################################################
class DummyPropagator:
    """Time propagator

    The DummyPropagator-class is the VIRTUAL base class for all propagators.

    """
    def __init__(self, td_density, td_hamiltonian, td_overlap,
                solver, preconditioner, gd, timer):
        """Create the DummyPropagator-object.

        Parameters
        ----------
        td_density: TimeDependentDensity
            the time-dependent density
        td_hamiltonian: TimeDependentHamiltonian
            the time-dependent hamiltonian
        td_overlap: TimeDependentOverlap
            the time-dependent overlap operator
        solver: LinearSolver
            solver for linear equations
        preconditioner: Preconditioner
            preconditioner for linear equations
        gd: GridDescriptor
            coarse (/wavefunction) grid descriptor
        timer: Timer
            timer

        """
        self.td_density = td_density
        self.td_hamiltonian = td_hamiltonian
        self.td_overlap = td_overlap

        self.wfs = td_density.get_wavefunctions()

        self.solver = solver
        self.preconditioner = preconditioner
        self.gd = gd
        self.timer = timer

        self.mblas = MultiBlas(gd)

    # Solve M psin = psi
    def apply_preconditioner(self, psi, psin):
        """Solves preconditioner equation.

        Parameters
        ----------
        psi: List of coarse grids
            the known wavefunctions
        psin: List of coarse grids
            the result

        """
        self.timer.start('Solve TDDFT preconditioner')
        if self.preconditioner is not None:
            self.preconditioner.apply(self.kpt, psi, psin)
        else:
            psin[:] = psi
        self.timer.stop('Solve TDDFT preconditioner')


    def propagate(self, time, time_step):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            the time step

        """
        raise RuntimeError('Error in DummyPropagator: '
                            'Member function propagate is virtual.')


###############################################################################
# ExplicitCrankNicolson
###############################################################################
class ExplicitCrankNicolson(DummyPropagator):
    """Explicit Crank-Nicolson propagator

    Crank-Nicolson propagator, which approximates the time-dependent
    Hamiltonian to be unchanged during one iteration step.

    (S(t) + .5j dt H(t) / hbar) psi(t+dt) = (S(t) - .5j dt H(t) / hbar) psi(t)

    """

    def __init__(self, td_density, td_hamiltonian, td_overlap,
                 solver, preconditioner, gd, timer):
        """Create ExplicitCrankNicolson-object.

        Parameters
        ----------
        td_density: TimeDependentDensity
            time-dependent density
        td_hamiltonian: TimeDependentHamiltonian
            time-dependent hamiltonian
        td_overlap: TimeDependentOverlap
            time-dependent overlap operator
        solver: LinearSolver
            solver for linear equations
        preconditioner: Preconditioner
            preconditioner for linear equations
        gd: GridDescriptor
            coarse (/wavefunction) grid descriptor
        timer: Timer
            timer

        """
        DummyPropagator.__init__(self, td_density, td_hamiltonian, td_overlap,
                                 solver, preconditioner, gd, timer)

        self.tmp_kpt_u = None
        self.hpsit = None
        self.spsit = None
        self.sinvhpsit = None

    # ( S + i H dt/2 ) psit(t+dt) = ( S - i H dt/2 ) psit(t)
    def propagate(self, time, time_step):
        """Propagate wavefunctions.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step

        """
        self.niter = 0

        # Allocate temporary wavefunctions
        if self.tmp_kpt_u is None:
            self.tmp_kpt_u = []
            for kpt in self.wfs.kpt_u:
                tmp_kpt = DummyKPoint()
                tmp_kpt.psit_nG = self.gd.empty(n=len(kpt.psit_nG),
                                                dtype=complex)
                self.tmp_kpt_u.append(tmp_kpt)

        # Allocate memory for Crank-Nicolson stuff
        nvec = len(self.wfs.kpt_u[0].psit_nG)
        if self.hpsit is None:
            self.hpsit = self.gd.zeros(nvec, dtype=complex)
        if self.spsit is None:
            self.spsit = self.gd.zeros(nvec, dtype=complex)


        self.timer.start('Update time-dependent operators')

        # Update overlap S(t) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t) based on the wavefunctions psit_nG
        # in kpt_u for t = time. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Update Hamiltonian H(t) to reflect density rho(t)
        self.td_hamiltonian.update(self.td_density.get_density(), time)

        self.timer.stop('Update time-dependent operators')

        # Copy current wavefunctions psit_nG to work wavefunction arrays
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG


        # Euler step
        # Overwrite psit_nG in tmp_kpt_u by (1 - i S^(-1)(t) H(t) dt) psit_nG
        # from corresponding kpt_u in a Euler step before predicting psit(t+dt)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.tmp_kpt_u):
            self.solve_propagation_equation(kpt, rhs_kpt, time_step, guess=True)

        # update projections before exiting
        self.td_overlap.update(self.wfs)

        return self.niter


    # ( S + i H dt/2 ) psit(t+dt) = ( S - i H dt/2 ) psit(t)
    def solve_propagation_equation(self, kpt, rhs_kpt, time_step, guess=False):

        # kpt is guess, rhs_kpt is used to calculate rhs and is overwritten
        nvec = len(rhs_kpt.psit_nG)

        assert kpt != rhs_kpt, 'Data race condition detected'
        assert len(kpt.psit_nG) == nvec, 'Incompatible lhs/rhs vectors'

        self.timer.start('Apply time-dependent operators')
        # Store H psi(t) as hpsit and S psit(t) as spsit
        self.td_overlap.update_k_point_projections(self.wfs, kpt, rhs_kpt.psit_nG)
        self.td_hamiltonian.apply(kpt, rhs_kpt.psit_nG, self.hpsit,
                                  calculate_P_ani=False)
        self.td_overlap.apply(rhs_kpt.psit_nG, self.spsit, self.wfs, kpt,
                              calculate_P_ani=False)
        self.timer.stop('Apply time-dependent operators')

        # Update rhs_kpt.psit_nG to reflect ( S - i H dt/2 ) psit(t)
        #rhs_kpt.psit_nG[:] = self.spsit - .5J * self.hpsit * time_step
        rhs_kpt.psit_nG[:] = self.spsit
        self.mblas.multi_zaxpy(-.5j*time_step, self.hpsit, rhs_kpt.psit_nG, nvec)

        if guess:
            if self.sinvhpsit is None:
                self.sinvhpsit = self.gd.zeros(len(kpt.psit_nG), dtype=complex)

            # Update estimate of psit(t+dt) to ( 1 - i S^(-1) H dt ) psit(t)
            self.td_overlap.apply_inverse(self.hpsit, self.sinvhpsit, self.wfs, kpt, use_cg=False)
            self.mblas.multi_zaxpy(-1.0j*time_step, self.sinvhpsit,
                                   kpt.psit_nG, nvec)

        # Information needed by solver.solve -> self.dot
        self.kpt = kpt
        self.time_step = time_step

        # Solve A x = b where A is (S + i H dt/2) and b = rhs_kpt.psit_nG
        self.niter += self.solver.solve(self, kpt.psit_nG, rhs_kpt.psit_nG)

    # ( S + i H dt/2 ) psi
    def dot(self, psi, psin):
        """Applies the propagator matrix to the given wavefunctions.

        Parameters
        ----------
        psi: List of coarse grids
            the known wavefunctions
        psin: List of coarse grids
            the result ( S + i H dt/2 ) psi

        """
        self.timer.start('Apply time-dependent operators')
        self.td_overlap.update_k_point_projections(self.wfs, self.kpt, psi)
        self.td_hamiltonian.apply(self.kpt, psi, self.hpsit,
                                  calculate_P_ani=False)
        self.td_overlap.apply(psi, self.spsit, self.wfs, self.kpt, calculate_P_ani=False)
        self.timer.stop('Apply time-dependent operators')

        # psin[:] = self.spsit + .5J * self.time_step * self.hpsit
        psin[:] = self.spsit
        self.mblas.multi_zaxpy(.5j*self.time_step, self.hpsit, psin, len(psi))



###############################################################################
# SemiImplicitCrankNicolson
###############################################################################
class SemiImplicitCrankNicolson(ExplicitCrankNicolson):
    """Semi-implicit Crank-Nicolson propagator

    Crank-Nicolson propagator, which first approximates the time-dependent
    Hamiltonian to be unchanged during one iteration step to predict future
    wavefunctions. Then the approximations for the future wavefunctions are
    used to approximate the Hamiltonian at the middle of the time step.

    (S(t) + .5j dt H(t) / hbar) psi(t+dt) = (S(t) - .5j dt H(t) / hbar) psi(t)
    (S(t) + .5j dt H(t+dt/2) / hbar) psi(t+dt)
    = (S(t) - .5j dt H(t+dt/2) / hbar) psi(t)

    """

    def __init__(self, td_density, td_hamiltonian, td_overlap,
                 solver, preconditioner, gd, timer):
        """Create SemiImplicitCrankNicolson-object.

        Parameters
        ----------
        td_density: TimeDependentDensity
            the time-dependent density
        td_hamiltonian: TimeDependentHamiltonian
            the time-dependent hamiltonian
        td_overlap: TimeDependentOverlap
            the time-dependent overlap operator
        solver: LinearSolver
            solver for linear equations
        preconditioner: Preconditioner
            preconditioner for linear equations
        gd: GridDescriptor
            coarse (wavefunction) grid descriptor
        timer: Timer
            timer

        """
        ExplicitCrankNicolson.__init__(self, td_density, td_hamiltonian,
                          td_overlap, solver, preconditioner, gd, timer)

        self.old_kpt_u = None


    def propagate(self, time, time_step):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step
        """

        self.niter = 0

        # Allocate old/temporary wavefunctions
        if self.old_kpt_u is None:
            self.old_kpt_u = []
            for kpt in self.wfs.kpt_u:
                old_kpt = DummyKPoint()
                old_kpt.psit_nG = self.gd.empty(n=len(kpt.psit_nG),
                                                dtype=complex)
                self.old_kpt_u.append(old_kpt)

        if self.tmp_kpt_u is None:
            self.tmp_kpt_u = []
            for kpt in self.wfs.kpt_u:
                tmp_kpt = DummyKPoint()
                tmp_kpt.psit_nG = self.gd.empty(n=len(kpt.psit_nG),
                                                dtype=complex)
                self.tmp_kpt_u.append(tmp_kpt)

        # Allocate memory for Crank-Nicolson stuff
        nvec = len(self.wfs.kpt_u[0].psit_nG)
        if self.hpsit is None:
            self.hpsit = self.gd.zeros(nvec, dtype=complex)
        if self.spsit is None:
            self.spsit = self.gd.zeros(nvec, dtype=complex)


        self.timer.start('Update time-dependent operators')

        # Update overlap S(t) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t) based on the wavefunctions psit_nG
        # in kpt_u for t = time. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Update Hamiltonian H(t) to reflect density rho(t)
        self.td_hamiltonian.update(self.td_density.get_density(), time)

        self.timer.stop('Update time-dependent operators')

        # Copy current wavefunctions psit_nG to work and old wavefunction arrays
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.old_kpt_u[u].psit_nG[:] = kpt.psit_nG
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG


        # Predictor step
        # Overwrite psit_nG in tmp_kpt_u by (1 - i S^(-1)(t) H(t) dt) psit_nG
        # from corresponding kpt_u in a Euler step before predicting psit(t+dt)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.tmp_kpt_u):
            self.solve_propagation_equation(kpt, rhs_kpt, time_step, guess=True)

        self.timer.start('Update time-dependent operators')

        # Update overlap S(t+dt) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t+dt) based on the wavefunctions psit_nG in
        # kpt_u for t = time+time_step. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Estimate Hamiltonian H(t+dt/2) by averaging H(t) and H(t+dt)
        # and retain the difference for a half-way Hamiltonian dH(t+dt/2).
        self.td_hamiltonian.half_update(self.td_density.get_density(),
                                        time + time_step)

        # Estimate overlap S(t+dt/2) by averaging S(t) and S(t+dt) #TODO!!!
        self.td_overlap.half_update(self.wfs)

        self.timer.stop('Update time-dependent operators')

        # Corrector step
        # Use predicted psit_nG in kpt_u as an initial guess, whereas the old
        # wavefunction in old_kpt_u are used to calculate rhs based on psit(t)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.old_kpt_u):
            # Average of psit(t) and predicted psit(t+dt)
            mean_psit_nG = 0.5*(kpt.psit_nG + rhs_kpt.psit_nG)
            self.td_hamiltonian.half_apply(kpt, mean_psit_nG, self.hpsit)
            self.td_overlap.apply_inverse(self.hpsit, self.sinvhpsit, self.wfs, kpt, use_cg=False)

            # Update kpt.psit_nG to reflect psit(t+dt) - i S^(-1) dH(t+dt/2) dt/2 psit(t+dt/2)
            kpt.psit_nG[:] = kpt.psit_nG - .5J * self.sinvhpsit * time_step
            self.mblas.multi_zaxpy(-.5j*time_step, self.sinvhpsit, kpt.psit_nG, nvec)

            self.solve_propagation_equation(kpt, rhs_kpt, time_step)

        # update projections before exiting
        self.td_overlap.update(self.wfs)

        return self.niter


    # ( S + i H dt/2 ) psit(t+dt) = ( S - i H dt/2 ) psit(t)
    def solve_propagation_equation(self, kpt, rhs_kpt, time_step, guess=False):

        # kpt is guess, rhs_kpt is used to calculate rhs and is overwritten
        nvec = len(rhs_kpt.psit_nG)

        assert kpt != rhs_kpt, 'Data race condition detected'
        assert len(kpt.psit_nG) == nvec, 'Incompatible lhs/rhs vectors'

        self.timer.start('Apply time-dependent operators')
        # Store H psi(t) as hpsit and S psit(t) as spsit
        self.td_overlap.update_k_point_projections(self.wfs, kpt, rhs_kpt.psit_nG)
        self.td_hamiltonian.apply(kpt, rhs_kpt.psit_nG, self.hpsit,
                                  calculate_P_ani=False)
        self.td_overlap.apply(rhs_kpt.psit_nG, self.spsit, self.wfs, kpt,
                              calculate_P_ani=False)
        self.timer.stop('Apply time-dependent operators')

        #self.mblas.multi_zdotc(self.shift, rhs_kpt.psit_nG, self.hpsit, nvec)
        #self.shift *= self.gd.dv
        #self.mblas.multi_zdotc(self.tmp_shift, rhs_kpt.psit_nG, self.spsit, nvec)
        #self.tmp_shift *= self.gd.dv
        #self.shift /= self.tmp_shift

        # Update rhs_kpt.psit_nG to reflect ( S - i H dt/2 ) psit(t)
        #rhs_kpt.psit_nG[:] = self.spsit - .5J * self.hpsit * time_step
        rhs_kpt.psit_nG[:] = self.spsit
        self.mblas.multi_zaxpy(-.5j*time_step, self.hpsit, rhs_kpt.psit_nG, nvec)
        # Apply shift -i eps S t/2
        #self.mblas.multi_zaxpy(-.5j*time_step * (-self.shift), self.spsit, rhs_kpt.psit_nG, nvec)

        if guess:
            if self.sinvhpsit is None:
                self.sinvhpsit = self.gd.zeros(len(kpt.psit_nG), dtype=complex)

            # Update estimate of psit(t+dt) to ( 1 - i S^(-1) H dt ) psit(t)
            self.td_overlap.apply_inverse(self.hpsit, self.sinvhpsit, self.wfs, kpt, use_cg=False)
            self.mblas.multi_zaxpy(-1.0j*time_step, self.sinvhpsit,
                                   kpt.psit_nG, nvec)

        # Information needed by solver.solve -> self.dot
        self.kpt = kpt
        self.time_step = time_step

        # Solve A x = b where A is (S + i H dt/2) and b = rhs_kpt.psit_nG
        self.niter += self.solver.solve(self, kpt.psit_nG, rhs_kpt.psit_nG)

        # Apply shift exp(i eps t)
        #self.phase_shift = np.exp(1.0J * self.shift * time_step)
        #self.mblas.multi_scale(self.phase_shift, kpt.psit_nG, nvec)

    # ( S + i H dt/2 ) psi
    def dot(self, psi, psin):
        """Applies the propagator matrix to the given wavefunctions.

        Parameters
        ----------
        psi: List of coarse grids
            the known wavefunctions
        psin: List of coarse grids
            the result ( S + i H dt/2 ) psi

        """
        ExplicitCrankNicolson.dot(self, psi, psin)
        # Apply shift -i eps S t/2
        #self.mblas.multi_zaxpy(.5j * self.time_step * (-self.shift), self.spsit, psin, len(psi))

class EhrenfestPAWSICN(ExplicitCrankNicolson):
    """Semi-implicit Crank-Nicolson propagator for Ehrenfest dynamics
       TODO: merge this with the ordinary SICN
    """
    def __init__(self, td_density, td_hamiltonian, td_overlap,
                 solver, preconditioner, gd, timer, corrector_guess = True,
                 predictor_guess = (True,False), use_cg = (False,False)):
        """Create SemiImplicitCrankNicolson-object.

        Parameters
        ----------
        td_density: TimeDependentDensity
            the time-dependent density
        td_hamiltonian: TimeDependentHamiltonian
            the time-dependent hamiltonian
        td_overlap: TimeDependentOverlap
            the time-dependent overlap operator
        solver: LinearSolver
            solver for linear equations
        preconditioner: Preconditioner
            preconditioner for linear equations
        gd: GridDescriptor
            coarse (wavefunction) grid descriptor
        timer: Timer
            timer
        corrector_guess: Bool
            use initial guess for the corrector step (default is True)
        predictor_guess: (Bool, Bool)
            use (first, second) order initial guesses for the predictor step
            default is (True, False)
        use_cg: (Bool, Bool)
            use CG for calculating the inverse overlap (predictor, corrector)
            default is (False, False)

        """
        ExplicitCrankNicolson.__init__(self, td_density, td_hamiltonian,
                          td_overlap, solver, preconditioner, gd, timer)

        self.old_kpt_u = None
        self.corrector_guess = corrector_guess
        self.predictor_guess = predictor_guess
        self.use_cg = use_cg

        #self.hsinvhpsit = None
        self.sinvh2psit = None

    def update_velocities(self, v_at_new, v_at_old = None):
        self.v_at = v_at_new.copy()
        if(v_at_old is not None):
            self.v_at_old = v_at_old.copy()

    def propagate(self, time, time_step, v_a):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step
        """

        self.niter = 0

        #update the atomic velocities which are required
        #for calculating the P term
        self.update_velocities(v_a)

        # Allocate old/temporary wavefunctions
        if self.old_kpt_u is None:
            self.old_kpt_u = []
            for kpt in self.wfs.kpt_u:
                old_kpt = DummyKPoint()
                old_kpt.psit_nG = self.gd.empty(n=len(kpt.psit_nG),
                                                dtype=complex)
                self.old_kpt_u.append(old_kpt)

        if self.tmp_kpt_u is None:
            self.tmp_kpt_u = []
            for kpt in self.wfs.kpt_u:
                tmp_kpt = DummyKPoint()
                tmp_kpt.psit_nG = self.gd.empty(n=len(kpt.psit_nG),
                                                dtype=complex)
                self.tmp_kpt_u.append(tmp_kpt)

        # Allocate memory for Crank-Nicolson stuff
        nvec = len(self.wfs.kpt_u[0].psit_nG)
        if self.hpsit is None:
            self.hpsit = self.gd.zeros(nvec, dtype=complex)
        if self.spsit is None:
            self.spsit = self.gd.zeros(nvec, dtype=complex)


        self.timer.start('Update time-dependent operators')

        # Update overlap S(t) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t) based on the wavefunctions psit_nG
        # in kpt_u for t = time. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Update Hamiltonian H(t) to reflect density rho(t)
        self.td_hamiltonian.update(self.td_density.get_density(), time)

        self.timer.stop('Update time-dependent operators')

        # Copy current wavefunctions psit_nG to work and old wavefunction arrays
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.old_kpt_u[u].psit_nG[:] = kpt.psit_nG
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG

        #print 'P_ani[0] =', self.wfs.kpt_u[0].P_ani[0]
        #print self.test


        # Predictor step
        # Overwrite psit_nG in tmp_kpt_u by (1 - i S^(-1)(t) H(t) dt) psit_nG
        # from corresponding kpt_u in a Euler step before predicting psit(t+dt)
        #self.v_at = self.v_at_old.copy() #v(t) for predictor step
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.tmp_kpt_u):
            #print 'self.predictor_guess[0]', self.predictor_guess[0]
            self.solve_propagation_equation(kpt, rhs_kpt, time_step, guess=self.predictor_guess[0])

        self.timer.start('Update time-dependent operators')

        # Update overlap S(t+dt) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t+dt) based on the wavefunctions psit_nG in
        # kpt_u for t = time+time_step. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Estimate Hamiltonian H(t+dt/2) by averaging H(t) and H(t+dt)
        # and retain the difference for a half-way Hamiltonian dH(t+dt/2).
        self.td_hamiltonian.half_update(self.td_density.get_density(),
                                        time + time_step)

        # Estimate overlap S(t+dt/2) by averaging S(t) and S(t+dt) #TODO!!!
        self.td_overlap.half_update(self.wfs)

        self.timer.stop('Update time-dependent operators')

        # Corrector step
        # Use predicted psit_nG in kpt_u as an initial guess, whereas the old
        # wavefunction in old_kpt_u are used to calculate rhs based on psit(t)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.old_kpt_u):
            # Average of psit(t) and predicted psit(t+dt)
            if(self.corrector_guess):
                mean_psit_nG = 0.5*(kpt.psit_nG + rhs_kpt.psit_nG)
                self.td_hamiltonian.half_apply(kpt, mean_psit_nG, self.hpsit)
                self.td_overlap.apply_inverse(self.hpsit, self.sinvhpsit, self.wfs, kpt, use_cg=self.use_cg[1])

                # Update kpt.psit_nG to reflect psit(t+dt) - i S^(-1) dH(t+dt/2) dt/2 psit(t+dt/2)
                kpt.psit_nG[:] = kpt.psit_nG - .5J * self.sinvhpsit * time_step
                self.mblas.multi_zaxpy(-.5j*time_step, self.sinvhpsit, kpt.psit_nG, nvec)

            self.solve_propagation_equation(kpt, rhs_kpt, time_step)

        # update projections before exiting
        self.td_overlap.update(self.wfs)

        return self.niter


    # ( S + i H dt/2 ) psit(t+dt) = ( S - i H dt/2 ) psit(t)
    def solve_propagation_equation(self, kpt, rhs_kpt, time_step, calculate_P_ani=False, guess=False):

        # kpt is guess, rhs_kpt is used to calculate rhs and is overwritten
        nvec = len(rhs_kpt.psit_nG)

        assert kpt != rhs_kpt, 'Data race condition detected'
        assert len(kpt.psit_nG) == nvec, 'Incompatible lhs/rhs vectors'

        self.timer.start('Apply time-dependent operators')
        # Store H psi(t) as hpsit and S psit(t) as spsit
        self.td_overlap.update_k_point_projections(self.wfs, kpt, rhs_kpt.psit_nG)
        self.td_hamiltonian.apply(kpt, rhs_kpt.psit_nG, self.hpsit,
                                  calculate_P_ani=False)
        self.td_hamiltonian.calculate_paw_correction(rhs_kpt.psit_nG, self.hpsit, self.wfs, kpt, self.v_at, calculate_P_ani=False)

        self.td_overlap.apply(rhs_kpt.psit_nG, self.spsit, self.wfs, kpt,
                              calculate_P_ani=False)
        self.timer.stop('Apply time-dependent operators')

        #self.mblas.multi_zdotc(self.shift, rhs_kpt.psit_nG, self.hpsit, nvec)
        #self.shift *= self.gd.dv
        #self.mblas.multi_zdotc(self.tmp_shift, rhs_kpt.psit_nG, self.spsit, nvec)
        #self.tmp_shift *= self.gd.dv
        #self.shift /= self.tmp_shift

        # Update rhs_kpt.psit_nG to reflect ( S - i H dt/2 ) psit(t)
        #rhs_kpt.psit_nG[:] = self.spsit - .5J * self.hpsit * time_step
        rhs_kpt.psit_nG[:] = self.spsit
        self.mblas.multi_zaxpy(-.5j*time_step, self.hpsit, rhs_kpt.psit_nG, nvec)
        # Apply shift -i eps S t/2
        #self.mblas.multi_zaxpy(-.5j*time_step * (-self.shift), self.spsit, rhs_kpt.psit_nG, nvec)

        if guess:
            if self.sinvhpsit is None:
                self.sinvhpsit = self.gd.zeros(len(kpt.psit_nG), dtype=complex)

            if self.predictor_guess[1]:
                if self.sinvh2psit is None:
                    self.sinvh2psit = self.gd.zeros(len(kpt.psit_nG), dtype=complex)

            # Update estimate of psit(t+dt) to ( 1 - i S^(-1) H dt ) psit(t)
            #print 'self.use_cg[0]', self.use_cg[0]
            self.td_overlap.apply_inverse(self.hpsit, self.sinvhpsit, self.wfs, kpt, use_cg=self.use_cg[0])
            #assert not self.use_cg[0]
            #self.td_overlap.apply_inverse(self.hpsit, self.sinvhpsit, self.wfs, kpt, use_cg=False)

            self.mblas.multi_zaxpy(-1.0j*time_step, self.sinvhpsit,
                                   kpt.psit_nG, nvec)
            #print 'using guess for P step'
            if(self.predictor_guess[1]):
                #print 'using 2nd order guess for P step'
                self.td_hamiltonian.apply(kpt, self.sinvhpsit, self.sinvh2psit,
                                  calculate_P_ani=False)
                self.td_hamiltonian.calculate_paw_correction(self.sinvhpsit, self.sinvh2psit, self.wfs, kpt, self.v_at, calculate_P_ani=False)
                self.td_overlap.apply_inverse(self.sinvh2psit, self.sinvh2psit, self.wfs, kpt, use_cg=self.use_cg[0])
                self.mblas.multi_zaxpy(-.5 * time_step*time_step, self.sinvh2psit,
                                   kpt.psit_nG, nvec)


        # Information needed by solver.solve -> self.dot
        self.kpt = kpt
        self.time_step = time_step

        # Solve A x = b where A is (S + i H dt/2) and b = rhs_kpt.psit_nG
        self.niter += self.solver.solve(self, kpt.psit_nG, rhs_kpt.psit_nG)

        # Apply shift exp(i eps t)
        #self.phase_shift = np.exp(1.0J * self.shift * time_step)
        #self.mblas.multi_scale(self.phase_shift, kpt.psit_nG, nvec)

    # ( S + i H dt/2 ) psi
    def dot(self, psi, psin):
        """Applies the propagator matrix to the given wavefunctions.

        Parameters
        ----------
        psi: List of coarse grids
            the known wavefunctions
        psin: List of coarse grids
            the result ( S + i H dt/2 ) psi

        """
        self.timer.start('Apply time-dependent operators')
        self.td_overlap.update_k_point_projections(self.wfs, self.kpt, psi)
        self.td_hamiltonian.apply(self.kpt, psi, self.hpsit,
                                  calculate_P_ani=False)
        self.td_hamiltonian.calculate_paw_correction(psi, self.hpsit, self.wfs, self.kpt, self.v_at, calculate_P_ani=False)
        self.td_overlap.apply(psi, self.spsit, self.wfs, self.kpt, calculate_P_ani=False)
        self.timer.stop('Apply time-dependent operators')

        # psin[:] = self.spsit + .5J * self.time_step * self.hpsit
        psin[:] = self.spsit
        self.mblas.multi_zaxpy(.5j*self.time_step, self.hpsit, psin, len(psi))
        # Apply shift -i eps S t/2
        #self.mblas.multi_zaxpy(.5j * self.time_step * (-self.shift), self.spsit, psin, len(psi))

class EhrenfestHGHSICN(ExplicitCrankNicolson):
    """Semi-implicit Crank-Nicolson propagator for Ehrenfest dynamics
       using HGH pseudopotentials

    """
    def __init__(self, td_density, td_hamiltonian, td_overlap,
                 solver, preconditioner, gd, timer):
        """Create SemiImplicitCrankNicolson-object.

        Parameters
        ----------
        td_density: TimeDependentDensity
            the time-dependent density
        td_hamiltonian: TimeDependentHamiltonian
            the time-dependent hamiltonian
        td_overlap: TimeDependentOverlap
            the time-dependent overlap operator
        solver: LinearSolver
            solver for linear equations
        preconditioner: Preconditioner
            preconditioner for linear equations
        gd: GridDescriptor
            coarse (wavefunction) grid descriptor
        timer: Timer
            timer

        """
        ExplicitCrankNicolson.__init__(self, td_density, td_hamiltonian,
                          td_overlap, solver, preconditioner, gd, timer)

        self.old_kpt_u = None

    def propagate(self, time, time_step):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step
        """

        self.niter = 0

        # Allocate old/temporary wavefunctions
        if self.old_kpt_u is None:
            self.old_kpt_u = []
            for kpt in self.wfs.kpt_u:
                old_kpt = DummyKPoint()
                old_kpt.psit_nG = self.gd.empty(n=len(kpt.psit_nG),
                                                dtype=complex)
                self.old_kpt_u.append(old_kpt)

        if self.tmp_kpt_u is None:
            self.tmp_kpt_u = []
            for kpt in self.wfs.kpt_u:
                tmp_kpt = DummyKPoint()
                tmp_kpt.psit_nG = self.gd.empty(n=len(kpt.psit_nG),
                                                dtype=complex)
                self.tmp_kpt_u.append(tmp_kpt)

        # Allocate memory for Crank-Nicolson stuff
        nvec = len(self.wfs.kpt_u[0].psit_nG)
        if self.hpsit is None:
            self.hpsit = self.gd.zeros(nvec, dtype=complex)
        if self.spsit is None:
            self.spsit = self.gd.zeros(nvec, dtype=complex)


        self.timer.start('Update time-dependent operators')

        # Update overlap S(t) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t) based on the wavefunctions psit_nG
        # in kpt_u for t = time. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Update Hamiltonian H(t) to reflect density rho(t)
        self.td_hamiltonian.update(self.td_density.get_density(), time)

        self.timer.stop('Update time-dependent operators')

        # Copy current wavefunctions psit_nG to work and old wavefunction arrays
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.old_kpt_u[u].psit_nG[:] = kpt.psit_nG
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG

        #print 'P_ani[0] =', self.wfs.kpt_u[0].P_ani[0]
        #print self.test


        # Predictor step
        # Overwrite psit_nG in tmp_kpt_u by (1 - i S^(-1)(t) H(t) dt) psit_nG
        # from corresponding kpt_u in a Euler step before predicting psit(t+dt)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.tmp_kpt_u):
            self.solve_propagation_equation(kpt, rhs_kpt, time_step, guess=False)

        self.timer.start('Update time-dependent operators')

        # Update overlap S(t+dt) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t+dt) based on the wavefunctions psit_nG in
        # kpt_u for t = time+time_step. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Estimate Hamiltonian H(t+dt/2) by averaging H(t) and H(t+dt)
        # and retain the difference for a half-way Hamiltonian dH(t+dt/2).
        self.td_hamiltonian.half_update(self.td_density.get_density(),
                                        time + time_step)

        # Estimate overlap S(t+dt/2) by averaging S(t) and S(t+dt) #TODO!!!
        self.td_overlap.half_update(self.wfs)

        self.timer.stop('Update time-dependent operators')

        # Corrector step
        # Use predicted psit_nG in kpt_u as an initial guess, whereas the old
        # wavefunction in old_kpt_u are used to calculate rhs based on psit(t)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.old_kpt_u):

            self.solve_propagation_equation(kpt, rhs_kpt, time_step)

        # update projections before exiting
        self.td_overlap.update(self.wfs)

        return self.niter


    # ( S + i H dt/2 ) psit(t+dt) = ( S - i H dt/2 ) psit(t)
    def solve_propagation_equation(self, kpt, rhs_kpt, time_step, calculate_P_ani = False, guess=False):

        # kpt is guess, rhs_kpt is used to calculate rhs and is overwritten
        nvec = len(rhs_kpt.psit_nG)

        assert kpt != rhs_kpt, 'Data race condition detected'
        assert len(kpt.psit_nG) == nvec, 'Incompatible lhs/rhs vectors'

        self.timer.start('Apply time-dependent operators')
        # Store H psi(t) as hpsit and S psit(t) as spsit
        self.td_overlap.update_k_point_projections(self.wfs, kpt, rhs_kpt.psit_nG)
        self.td_hamiltonian.apply(kpt, rhs_kpt.psit_nG, self.hpsit,
                                  calculate_P_ani=False)
        self.td_overlap.apply(rhs_kpt.psit_nG, self.spsit, self.wfs, kpt,
                              calculate_P_ani=False)
        self.timer.stop('Apply time-dependent operators')

        # Update rhs_kpt.psit_nG to reflect ( S - i H dt/2 ) psit(t)
        #rhs_kpt.psit_nG[:] = self.spsit - .5J * self.hpsit * time_step
        rhs_kpt.psit_nG[:] = self.spsit
        self.mblas.multi_zaxpy(-.5j*time_step, self.hpsit, rhs_kpt.psit_nG, nvec)
        # Apply shift -i eps S t/2
        #self.mblas.multi_zaxpy(-.5j*time_step * (-self.shift), self.spsit, rhs_kpt.psit_nG, nvec)

        # Information needed by solver.solve -> self.dot
        self.kpt = kpt
        self.time_step = time_step

        # Solve A x = b where A is (S + i H dt/2) and b = rhs_kpt.psit_nG
        self.niter += self.solver.solve(self, kpt.psit_nG, rhs_kpt.psit_nG)

        # Apply shift exp(i eps t)
        #self.phase_shift = np.exp(1.0J * self.shift * time_step)
        #self.mblas.multi_scale(self.phase_shift, kpt.psit_nG, nvec)

    # ( S + i H dt/2 ) psi
    def dot(self, psi, psin):
        """Applies the propagator matrix to the given wavefunctions.

        Parameters
        ----------
        psi: List of coarse grids
            the known wavefunctions
        psin: List of coarse grids
            the result ( S + i H dt/2 ) psi

        """
        ExplicitCrankNicolson.dot(self, psi, psin)





###############################################################################
# EnforcedTimeReversalSymmetryCrankNicolson
###############################################################################
class EnforcedTimeReversalSymmetryCrankNicolson(ExplicitCrankNicolson):
    """Enforced time-reversal symmetry Crank-Nicolson propagator

    Crank-Nicolson propagator, which first approximates the time-dependent
    Hamiltonian to be unchanged during one iteration step to predict future
    wavefunctions. Then the approximations for the future wavefunctions are
    used to approximate the Hamiltonian in the future.

    (S(t) + .5j dt H(t) / hbar) psi(t+dt) = (S(t) - .5j dt H(t) / hbar) psi(t)
    (S(t) + .5j dt H(t+dt) / hbar) psi(t+dt)
    = (S(t) - .5j dt H(t) / hbar) psi(t)

    """

    def __init__(self, td_density, td_hamiltonian, td_overlap,
                 solver, preconditioner, gd, timer):
        """Create SemiImplicitCrankNicolson-object.

        Parameters
        ----------
        td_density: TimeDependentDensity
            the time-dependent density
        td_hamiltonian: TimeDependentHamiltonian
            the time-dependent hamiltonian
        td_overlap: TimeDependentOverlap
            the time-dependent overlap operator
        solver: LinearSolver
            solver for linear equations
        preconditioner: Preconditioner
            preconditioner for linear equations
        gd: GridDescriptor
            coarse (wavefunction) grid descriptor
        timer: Timer
            timer

        """
        ExplicitCrankNicolson.__init__(self, td_density, td_hamiltonian,
                          td_overlap, solver, preconditioner, gd, timer)

        self.old_kpt_u = None


    def propagate(self, time, time_step, update_callback=None):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step
        """

        self.niter = 0

        # Allocate old/temporary wavefunctions
        if self.old_kpt_u is None:
            self.old_kpt_u = []
            for kpt in self.wfs.kpt_u:
                old_kpt = DummyKPoint()
                old_kpt.psit_nG = self.gd.empty(n=len(kpt.psit_nG),
                                                dtype=complex)
                self.old_kpt_u.append(old_kpt)

        if self.tmp_kpt_u is None:
            self.tmp_kpt_u = []
            for kpt in self.wfs.kpt_u:
                tmp_kpt = DummyKPoint()
                tmp_kpt.psit_nG = self.gd.empty(n=len(kpt.psit_nG),
                                                dtype=complex)
                self.tmp_kpt_u.append(tmp_kpt)

        # Allocate memory for Crank-Nicolson stuff
        nvec = len(self.wfs.kpt_u[0].psit_nG)
        if self.hpsit is None:
            self.hpsit = self.gd.zeros(nvec, dtype=complex)
        if self.spsit is None:
            self.spsit = self.gd.zeros(nvec, dtype=complex)


        self.timer.start('Update time-dependent operators')

        # Update overlap S(t) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t) based on the wavefunctions psit_nG
        # in kpt_u for t = time. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Update Hamiltonian H(t) to reflect density rho(t)
        self.td_hamiltonian.update(self.td_density.get_density(), time)

        self.timer.stop('Update time-dependent operators')

        # Copy current wavefunctions psit_nG to work and old wavefunction arrays
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.old_kpt_u[u].psit_nG[:] = kpt.psit_nG
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG


        # Predictor step
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.old_kpt_u):
            self.create_rhs(rhs_kpt, kpt, time_step)

        if update_callback is not None:
            update_callback()


        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.old_kpt_u):
            self.solve_propagation_equation(kpt, rhs_kpt, time_step)


        self.timer.start('Update time-dependent operators')

        # Update overlap S(t+dt) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t+dt) based on the wavefunctions psit_nG in
        # kpt_u for t = time+time_step. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Estimate Hamiltonian H(t+dt/2) by averaging H(t) and H(t+dt)
        # and retain the difference for a half-way Hamiltonian dH(t+dt/2).
        self.td_hamiltonian.update(self.td_density.get_density(),
                                   time + time_step)

        # Estimate overlap S(t+dt/2) by averaging S(t) and S(t+dt) #TODO!!!
        self.td_overlap.update(self.wfs)

        self.timer.stop('Update time-dependent operators')

        # Corrector step
        # Use predicted psit_nG in kpt_u as an initial guess, whereas the old
        # wavefunction in old_kpt_u are used to calculate rhs based on psit(t)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.old_kpt_u):
            self.solve_propagation_equation(kpt, rhs_kpt, time_step)

        # update projections before exiting
        self.td_overlap.update(self.wfs)

        return self.niter

    # Create RHS
    def create_rhs(self, rhs_kpt, kpt, time_step):
        # kpt is guess, rhs_kpt is used to calculate rhs and is overwritten
        nvec = len(rhs_kpt.psit_nG)

        assert kpt != rhs_kpt, 'Data race condition detected'
        assert len(kpt.psit_nG) == nvec, 'Incompatible lhs/rhs vectors'

        self.timer.start('Apply time-dependent operators')
        # Store H psi(t) as hpsit and S psit(t) as spsit
        self.td_overlap.update_k_point_projections(self.wfs, kpt, rhs_kpt.psit_nG)
        self.td_hamiltonian.apply(kpt, rhs_kpt.psit_nG, self.hpsit,
                                  calculate_P_ani=False)
        self.td_overlap.apply(rhs_kpt.psit_nG, self.spsit, self.wfs, kpt,
                              calculate_P_ani=False)
        self.timer.stop('Apply time-dependent operators')

        # Update rhs_kpt.psit_nG to reflect ( S - i H dt/2 ) psit(t)
        #rhs_kpt.psit_nG[:] = self.spsit - .5J * self.hpsit * time_step
        rhs_kpt.psit_nG[:] = self.spsit
        self.mblas.multi_zaxpy(-.5j*time_step, self.hpsit, rhs_kpt.psit_nG, nvec)

    # ( S + i H(t+dt) dt/2 ) psit(t+dt) = ( S - i H(t) dt/2 ) psit(t)
    # rhs_kpt = ( S - i H(t) dt/2 ) psit(t)
    def solve_propagation_equation(self, kpt, rhs_kpt, time_step, guess=False):
        # Information needed by solver.solve -> self.dot
        self.kpt = kpt
        self.time_step = time_step

        # Solve A x = b where A is (S + i H dt/2) and b = rhs_kpt.psit_nG
        self.niter += self.solver.solve(self, kpt.psit_nG, rhs_kpt.psit_nG)

        # Apply shift exp(i eps t)
        #self.phase_shift = np.exp(1.0J * self.shift * time_step)
        #self.mblas.multi_scale(self.phase_shift, kpt.psit_nG, nvec)

    # ( S + i H dt/2 ) psi
    def dot(self, psi, psin):
        """Applies the propagator matrix to the given wavefunctions.

        Parameters
        ----------
        psi: List of coarse grids
            the known wavefunctions
        psin: List of coarse grids
            the result ( S + i H dt/2 ) psi

        """
        ExplicitCrankNicolson.dot(self, psi, psin)
        # Apply shift -i eps S t/2
        #self.mblas.multi_zaxpy(.5j * self.time_step * (-self.shift), self.spsit, psin, len(psi))






###############################################################################
# AbsorptionKick
###############################################################################
class AbsorptionKick(ExplicitCrankNicolson):
    """Absorption kick propagator

    Absorption kick propagator::

      (S(t) + .5j dt p.r / hbar) psi(0+) = (S(t) - .5j dt p.r / hbar) psi(0-)

    where ``|p| = (eps e / hbar)``, and eps is field strength, e is elementary
    charge.

    """

    def __init__(self, wfs, abs_kick_hamiltonian, td_overlap,
                 solver, preconditioner, gd, timer):
        """Create AbsorptionKick-object.

        Parameters
        ----------
        wfs: FDWaveFunctions
            time-independent grid-based wavefunctions
        abs_kick_hamiltonian: AbsorptionKickHamiltonian
            the absorption kick hamiltonian
        td_overlap: TimeDependentOverlap
            the time-dependent overlap operator
        solver: LinearSolver
            solver for linear equations
        preconditioner: Preconditioner
            preconditioner for linear equations
        gd: GridDescriptor
            coarse (wavefunction) grid descriptor
        timer: Timer
            timer

        """
        ExplicitCrankNicolson.__init__(self, DummyDensity(wfs),
                        abs_kick_hamiltonian, td_overlap, solver,
                        preconditioner, gd, timer)


    def kick(self):
        """Excite all possible frequencies.

        """

        # if rank == 0:
        #     self.text('Kick iterations = ', self.td_hamiltonian.iterations)

        for l in range(self.td_hamiltonian.iterations):
            self.propagate(0, 1.0)
            # if rank == 0:
            #     self.text('.')
        # if rank == 0:
        #     print ''






###############################################################################
# SemiImpicitTaylorExponential
###############################################################################
class SemiImplicitTaylorExponential(DummyPropagator):
    """Semi-implicit Taylor exponential propagator
    exp(-i S^-1 H t) = 1 - i S^-1 H t + (1/2) (-i S^-1 H t)^2 + ...

    """

    def __init__(self, td_density, td_hamiltonian, td_overlap, solver,
                 preconditioner, gd, timer, degree=4):
        """Create SemiImplicitTaylorExponential-object.

        Parameters
        ----------
        td_density: TimeDependentDensity
            the time-dependent density
        td_hamiltonian: TimeDependentHamiltonian
            the time-dependent hamiltonian
        td_overlap: TimeDependentOverlap
            the time-dependent overlap operator
        solver: LinearSolver
            solver for linear equations
        preconditioner: Preconditioner
            preconditioner
        gd: GridDescriptor
            coarse (wavefunction) grid descriptor
        timer: Timer
            timer
        degree: integer
            Degree of the Taylor polynomial (default is 4)

        """
        DummyPropagator.__init__(self, td_density, td_hamiltonian, td_overlap,
                                 solver, preconditioner, gd, timer)

        self.degree = degree

        self.tmp_kpt_u = None
        self.psin = None
        self.hpsit = None


    def propagate(self, time, time_step):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step
        """

        self.niter = 0

        # Allocate temporary wavefunctions
        if self.tmp_kpt_u is None:
            self.tmp_kpt_u = []
            for kpt in self.wfs.kpt_u:
                tmp_kpt = DummyKPoint()
                tmp_kpt.psit_nG = self.gd.empty(n=len(kpt.psit_nG),
                                                dtype=complex)
                self.tmp_kpt_u.append(tmp_kpt)

        # Allocate memory for Taylor exponential stuff
        nvec = len(self.wfs.kpt_u[0].psit_nG)
        if self.psin is None:
            self.psin = self.gd.zeros(nvec, dtype=complex)
        if self.hpsit is None:
            self.hpsit = self.gd.zeros(nvec, dtype=complex)


        self.timer.start('Update time-dependent operators')

        # Update overlap S(t) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t) based on the wavefunctions psit_nG
        # in kpt_u for t = time. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Update Hamiltonian H(t) to reflect density rho(t)
        self.td_hamiltonian.update(self.td_density.get_density(), time)

        self.timer.stop('Update time-dependent operators')


        # copy current wavefunctions to temporary variable
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG

        # predict for each k-point
        for kpt in self.wfs.kpt_u:
            self.solve_propagation_equation(kpt, time_step)


        self.timer.start('Update time-dependent operators')

        # Update overlap S(t+dt) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t+dt) based on the wavefunctions psit_nG in
        # kpt_u for t = time+time_step. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Estimate Hamiltonian H(t+dt/2) by averaging H(t) and H(t+dt)
        self.td_hamiltonian.half_update(self.td_density.get_density(),
                                        time + time_step)

        # Estimate overlap S(t+dt/2) by averaging S(t) and S(t+dt) #TODO!!!
        self.td_overlap.half_update(self.wfs)

        self.timer.stop('Update time-dependent operators')


        # propagate psit(t), not psit(t+dt), in correct
        for u, kpt in enumerate(self.wfs.kpt_u):
            kpt.psit_nG[:] = self.tmp_kpt_u[u].psit_nG

        # correct for each k-point
        for kpt in self.wfs.kpt_u:
            self.solve_propagation_equation(kpt, time_step)

        # update projections before exiting
        self.td_overlap.update(self.wfs)

        return self.niter

    # psi(t) = exp(-i t S^-1 H) psi(0)
    # psi(t) = 1  + (-i S^-1 H t) (1 + (1/2) (-i S^-1 H t) (1 + ... ) )
    def solve_propagation_equation(self, kpt, time_step):

        nvec = len(kpt.psit_nG)

        # Information needed by solver.solve -> self.dot
        self.kpt = kpt
        self.time_step = time_step

        # psin = psi(0)
        self.psin[:] = kpt.psit_nG
        for k in range(self.degree,0,-1):
            # psin = psi(0) + (1/k) (-i S^-1 H t) psin
            self.td_hamiltonian.apply(kpt, self.psin, self.hpsit)
            # S psin = H psin
            self.psin[:] = self.hpsit
            self.niter += self.solver.solve(self, self.psin, self.hpsit)
            #print 'Linear solver iterations = ', self.solver.iterations
            # psin = psi(0) + (-it/k) S^-1 H psin
            self.mblas.multi_scale(-1.0j*time_step/k, self.psin, nvec)
            self.mblas.multi_zaxpy(1.0, kpt.psit_nG, self.psin, nvec)

        kpt.psit_nG[:] = self.psin

    def dot(self, psit, spsit):
        self.td_overlap.apply(psit, spsit, self.wfs, self.kpt)




###############################################################################
# SemiImplicitKrylovExponential
###############################################################################
class SemiImplicitKrylovExponential(DummyPropagator):
    """Semi-implicit Krylov exponential propagator


    """

    def __init__(self, td_density, td_hamiltonian, td_overlap, solver,
                 preconditioner, gd, timer, degree=4):
        """Create SemiImplicitKrylovExponential-object.

        Parameters
        ----------
        td_density: TimeDependentDensity
            the time-dependent density
        td_hamiltonian: TimeDependentHamiltonian
            the time-dependent hamiltonian
        td_overlap: TimeDependentOverlap
            the time-dependent overlap operator
        solver: LinearSolver
            solver for linear equations
        preconditioner: Preconditioner
            preconditioner
        gd: GridDescriptor
            coarse (wavefunction) grid descriptor
        timer: Timer
            timer
        degree: integer
            Degree of the Krylov subspace (default is 4)

        """
        DummyPropagator.__init__(self, td_density, td_hamiltonian, td_overlap,
                                 solver, preconditioner, gd, timer)

        self.kdim = degree + 1

        self.tmp_kpt_u = None
        self.lm = None
        self.em = None
        self.hm = None
        self.sm = None
        self.xm = None
        self.qm = None
        self.Hqm = None
        self.Sqm = None
        self.rqm = None


    def propagate(self, time, time_step):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step
        """

        self.niter = 0

        # Allocate temporary wavefunctions
        if self.tmp_kpt_u is None:
            self.tmp_kpt_u = []
            for kpt in self.wfs.kpt_u:
                tmp_kpt = DummyKPoint()
                tmp_kpt.psit_nG = self.gd.empty(n=len(kpt.psit_nG),
                                                dtype=complex)
                self.tmp_kpt_u.append(tmp_kpt)


        # Allocate memory for Krylov subspace stuff
        nvec = len(self.wfs.kpt_u[0].psit_nG)

        # em = (wfs, degree)
        if self.em is None:
            self.em = np.zeros((nvec, self.kdim), float)

        # lm = (wfs)
        if self.lm is None:
            self.lm = np.zeros((nvec,), complex)

        # hm = (wfs, degree, degree)
        if self.hm is None:
            self.hm = np.zeros((nvec, self.kdim, self.kdim), complex)
        # sm = (wfs, degree, degree)
        if self.sm is None:
            self.sm = np.zeros((nvec, self.kdim, self.kdim), complex)
        # xm = (wfs, degree, degree)
        if self.xm is None:
            self.xm = np.zeros((nvec, self.kdim, self.kdim), complex)

        # qm = (degree, wfs, nx, ny, nz)
        if self.qm is None:
            self.qm = self.gd.zeros((self.kdim, nvec), dtype=complex)
        # H qm = (degree, wfs, nx, ny, nz)
        if self.Hqm is None:
            self.Hqm = self.gd.zeros((self.kdim, nvec), dtype=complex)
        # S qm = (degree, wfs, nx, ny, nz)
        if self.Sqm is None:
            self.Sqm = self.gd.zeros((self.kdim, nvec), dtype=complex)
        # rqm = (wfs, nx, ny, nz)
        if self.rqm is None:
            self.rqm = self.gd.zeros((nvec,), dtype=complex)


        self.time_step = time_step


        self.timer.start('Update time-dependent operators')

        # Update overlap S(t) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t) based on the wavefunctions psit_nG
        # in kpt_u for t = time. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Update Hamiltonian H(t) to reflect density rho(t)
        self.td_hamiltonian.update(self.td_density.get_density(), time)

        self.timer.stop('Update time-dependent operators')


        # copy current wavefunctions to temporary variable
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG

        # predict for each k-point
        for kpt in self.wfs.kpt_u:
            self.solve_propagation_equation(kpt, time_step)


        self.timer.start('Update time-dependent operators')

        # Update overlap S(t+dt) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t+dt) based on the wavefunctions psit_nG in
        # kpt_u for t = time+time_step. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Estimate Hamiltonian H(t+dt/2) by averaging H(t) and H(t+dt)
        self.td_hamiltonian.half_update(self.td_density.get_density(),
                                        time + time_step)

        # Estimate overlap S(t+dt/2) by averaging S(t) and S(t+dt) #TODO!!!
        self.td_overlap.half_update(self.wfs)

        self.timer.stop('Update time-dependent operators')


        # propagate psit(t), not psit(t+dt), in correct
        for u, kpt in enumerate(self.wfs.kpt_u):
            kpt.psit_nG[:] = self.tmp_kpt_u[u].psit_nG

        # correct for each k-point
        for kpt in self.wfs.kpt_u:
            self.solve_propagation_equation(kpt, time_step)

        # update projections before exiting
        self.td_overlap.update(self.wfs)

        return self.niter

    # psi(t) = exp(-i t S^-1 H) psi(0)
    def solve_propagation_equation(self, kpt, time_step):

        nvec = len(kpt.psit_nG)
        tmp = np.zeros((nvec,), complex)
        xm_tmp = np.zeros((nvec, self.kdim), complex)

        qm = self.qm
        Hqm = self.Hqm
        Sqm = self.Sqm

        # Information needed by solver.solve -> self.dot
        self.kpt = kpt

        scale = self.create_krylov_subspace(kpt,
                                            self.td_hamiltonian,
                                            self.td_overlap,
                                            self.qm, self.Hqm, self.Sqm)

        # Calculate hm and sm
        for i in range(self.kdim):
            for j in range(self.kdim):
                self.mblas.multi_zdotc(tmp, qm[i], Hqm[j], nvec)
                tmp *= self.gd.dv
                for k in range(nvec):
                    self.hm[k][i][j] = tmp[k]
                self.mblas.multi_zdotc(tmp, qm[i], Sqm[j], nvec)
                tmp *= self.gd.dv
                for k in range(nvec):
                    self.sm[k][i][j] = tmp[k]

        #print 'Hm ='
        #print np.round(self.hm*1e4) / 1e4
        #print 'log Hm ='
        #print np.round(np.log(self.hm)*1e2)/1e2
        #print 'Sm ='
        #print np.round(self.sm*1e4) / 1e4

        #print np.round(np.log10(np.abs(np.linalg.eigh(self.hm[0])[1]))*1e6)/1e6

        # Diagonalize
        # Propagate
        # psi(t) = Qm Xm exp(-i Em t) Xm^H Sm e_1
        #        = Qm Xm exp(-i Em t) Sm Qm^H S psi(0) ???
        #        = Qm Xm exp(-i Em t) y
        #        = Qm Xm z
        # y = Sm Qm^H S psi(0) = Xm^H Sm e_1
        # if Sm = I then y is the first row of Xm^*
        # and z = exp(-i Em t) y
        for k in range(nvec):
            (self.em[k], self.xm[k]) = np.linalg.eigh(self.hm[k])
        #print 'Em = ', self.em
        #for k in range(nvec):
            #print 'Xm',k,' = '
            #print self.xm[k]

        #print self.em[0] * (-1.0J*self.time_step)
        self.em = np.exp(self.em * (-1.0j*time_step))
        #print self.em[0]
        #print np.linalg.eigh(self.hm[0])
        for k in range(nvec):
            z = self.em[k] * np.conj(self.xm[k,0])
            xm_tmp[k][:] = np.dot(self.xm[k], z)
        #print xm_tmp
        kpt.psit_nG[:] = 0.0
        for k in range(nvec):
            for i in range(self.kdim):
                #print 'Xm_tmp[',k,'][',i,'] = ', xm_tmp[k][i]
                axpy(xm_tmp[k][i] / scale[k],
                     self.qm[i][k], kpt.psit_nG[k])

        #print self.qm
        #print kpt.psit_nG

    # Create Krylov subspace
    #    K_v = { psi, S^-1 H psi, (S^-1 H)^2 psi, ... }

    def create_krylov_subspace(self, kpt, h, s, qm, Hqm, Sqm):
        nvec = len(kpt.psit_nG)
        # tmp = (wfs)
        tmp = np.zeros((nvec,), complex)
        scale = np.zeros((nvec,), complex)
        scale[:] = 0.0
        rqm = self.rqm

        # q_0 = psi
        rqm[:] = kpt.psit_nG

        for i in range(self.kdim):
            qm[i][:] = rqm

            # S orthogonalize
            # q_i = q_i - sum_j<i <q_j|S|q_i> q_j
            for j in range(i):
                self.mblas.multi_zdotc(tmp, qm[i], Sqm[j], nvec)
                tmp *= self.gd.dv
                tmp = np.conj(tmp)
                self.mblas.multi_zaxpy(-tmp, qm[j], qm[i], nvec)

            # S q_i
            s.apply(qm[i], Sqm[i], self.wfs, kpt)
            self.mblas.multi_zdotc(tmp, qm[i], Sqm[i], nvec)
            tmp *= self.gd.dv
            self.mblas.multi_scale(1./np.sqrt(tmp), qm[i], nvec)
            self.mblas.multi_scale(1./np.sqrt(tmp), Sqm[i], nvec)
            if i == 0:
                scale[:] = 1/np.sqrt(tmp)
                #print 'Scale', scale

            # H q_i
            h.apply(kpt, qm[i], Hqm[i])

            # S r = H q_i, (if stuff, to save one inversion)
            if i+1 < self.kdim:
                rqm[:] = Hqm[i]
                self.solver.solve(self, rqm, Hqm[i])
                #print 'Linear solver iterations = ', self.solver.iterations

        #print '---'
        return scale

    def dot(self, psit, spsit):
        self.td_overlap.apply(psit, spsit, self.wfs, self.kpt)

    ### Below this, just for testing & debug
    def Sdot(self, psit, spsit):
        self.apply_preconditioner(psit, self.tmp)
        self.td_overlap.apply(self.tmp, spsit, self.wfs, self.kpt)

    def Hdot(self, psit, spsit):
        self.apply_preconditioner(psit, self.tmp)
        self.td_hamiltonian.apply(self.kpt, self.tmp, spsit)

    def inverse_overlap(self, kpt_u, degree):
        self.dot = self.Sdot
        self.kpt = kpt_u[0]
        nvec = len(self.kpt.psit_nG)
        nrm2 = np.zeros(nvec, dtype=complex)
        self.tmp = self.gd.zeros(n=nvec, dtype=complex)

        for i in range(10):
            self.solver.solve(self, self.kpt.psit_nG, self.kpt.psit_nG)
            self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.kpt.psit_nG, nvec)
            nrm2 *= self.gd.dv
            self.mblas.multi_scale(1/np.sqrt(nrm2), self.kpt.psit_nG, nvec)
        self.td_overlap.apply(self.kpt.psit_nG, self.tmp, self.wfs, self.kpt)
        self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.tmp, nvec)
        nrm2 *= self.gd.dv
        print('S min eig = ', nrm2)


    def overlap(self, kpt_u, degree):
        self.dot = self.Sdot
        self.kpt = kpt_u[0]
        nvec = len(self.kpt.psit_nG)
        nrm2 = np.zeros(nvec, dtype=complex)
        self.tmp = self.gd.zeros(n=nvec, dtype=complex)

        for i in range(100):
            self.tmp[:] = self.kpt.psit_nG
            self.td_overlap.apply(self.tmp, self.kpt.psit_nG, self.wfs, self.kpt)
            self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.kpt.psit_nG, nvec)
            nrm2 *= self.gd.dv
            self.mblas.multi_scale(1/np.sqrt(nrm2), self.kpt.psit_nG, nvec)
        self.td_overlap.apply(self.kpt.psit_nG, self.tmp, self.wfs, self.kpt)
        self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.tmp, nvec)
        nrm2 *= self.gd.dv
        print('S max eig = ', nrm2)


    def inverse_hamiltonian(self, kpt_u, degree):
        self.dot = self.Hdot
        self.kpt = kpt_u[0]
        nvec = len(self.kpt.psit_nG)
        nrm2 = np.zeros(nvec, dtype=complex)
        self.tmp = self.gd.zeros(n=nvec, dtype=complex)

        for i in range(10):
            self.solver.solve(self, self.kpt.psit_nG, self.kpt.psit_nG)
            self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.kpt.psit_nG, nvec)
            nrm2 *= self.gd.dv
            self.mblas.multi_scale(1/np.sqrt(nrm2), self.kpt.psit_nG, nvec)
        self.td_hamiltonian.apply(self.kpt, self.kpt.psit_nG, self.tmp)
        self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.tmp, nvec)
        nrm2 *= self.gd.dv
        print('H min eig = ', nrm2)


    def hamiltonian(self, kpt_u, degree):
        self.dot = self.Hdot
        self.kpt = kpt_u[0]
        nvec = len(self.kpt.psit_nG)
        nrm2 = np.zeros(nvec, dtype=complex)
        self.tmp = self.gd.zeros(n=nvec, dtype=complex)

        for i in range(100):
            self.tmp[:] = self.kpt.psit_nG
            self.td_hamiltonian.apply(self.kpt, self.tmp, self.kpt.psit_nG)
            self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.kpt.psit_nG, nvec)
            nrm2 *= self.gd.dv
            self.mblas.multi_scale(1/np.sqrt(nrm2), self.kpt.psit_nG, nvec)
        self.td_hamiltonian.apply(self.kpt, self.kpt.psit_nG, self.tmp)
        self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.tmp, nvec)
        nrm2 *= self.gd.dv
        print('H max eig = ', nrm2)



from ase import Atom, Atoms
from gpaw import GPAW
from gpaw.test import equal

a = 4.05
d = a / 2**0.5
bulk = Atoms([Atom('Al', (0, 0, 0)),
              Atom('Al', (0.5, 0.5, 0.5))], pbc=True)
bulk.set_cell((d, d, a), scale_atoms=True)
h = 0.25
calc = GPAW(mode='pw',
            nbands=2 * 8,
            kpts=(2, 2, 2),
            convergence={'eigenstates': 7.2e-9, 'energy': 1e-5})
bulk.set_calculator(calc)
e0 = bulk.get_potential_energy()
niter0 = calc.get_number_of_iterations()
calc = GPAW(mode='pw',
            nbands=2 * 8,
            kpts=(2, 2, 2),
            convergence={'eigenstates': 7.2e-9,
                         'energy': 1e-5,
                         'bands': 5},
            eigensolver='dav')
bulk.set_calculator(calc)
e1 = bulk.get_potential_energy()
niter1 = calc.get_number_of_iterations()
equal(e0, e1, 5.0e-6)

energy_tolerance = 0.0004
niter_tolerance = 0
equal(e0, -6.97798, energy_tolerance)
assert 8 <= niter0 <= 12, niter0
equal(e1, -6.97798, energy_tolerance)
assert 8 <= niter1 <= 22, niter1



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""Occupation number objects."""

import warnings
from math import pi

import numpy as np
from ase.units import Hartree

from gpaw.utilities import erf
from gpaw.mpi import serial_comm


def create_occupation_number_object(name, **kwargs):
    if name == 'fermi-dirac':
        return FermiDirac(**kwargs)
    if name == 'methfessel-paxton':
        return MethfesselPaxton(**kwargs)
    if name == 'orbital-free':
        return TFOccupations()
    raise ValueError('Unknown occupation number object name: ' + name)


def occupation_numbers(occ, eps_skn, weight_k, nelectrons):
    """Calculate occupation numbers from eigenvalues in eV.

    occ: dict
        Example: {'name': 'fermi-dirac', 'width': 0.05} (width in eV).
    eps_skn: ndarray, shape=(nspins, nibzkpts, nbands)
        Eigenvalues.
    weight_k: ndarray, shape=(nibzkpts,)
        Weights of k-points in IBZ (must sum to 1).
    nelectrons: int or float
        Number of electrons.

    Returns a tuple containing:

    * f_skn (sums to nelectrons)
    * fermi-level
    * magnetic moment
    * entropy as -S*T
    """

    from gpaw.grid_descriptor import GridDescriptor

    occ = create_occupation_number_object(**occ)

    eps_skn = np.asarray(eps_skn) / Hartree
    weight_k = np.asarray(weight_k)
    nspins, nkpts, nbands = eps_skn.shape
    f_skn = np.empty_like(eps_skn)

    class SimpleNamespace:
        """Same as types.SimpleNamespace from Python 3.3+."""
        def __init__(self, **kwargs):
            self.__dict__.update(kwargs)

    wfs = SimpleNamespace(kpt_u=[],
                          nvalence=nelectrons,
                          nspins=nspins,
                          kptband_comm=serial_comm,
                          world=serial_comm,
                          gd=GridDescriptor([4, 4, 4], [1.0, 1.0, 1.0],
                                            comm=serial_comm),
                          bd=SimpleNamespace(nbands=nbands,
                                             collect=lambda x: x,
                                             comm=serial_comm),
                          kd=SimpleNamespace(mynks=nspins * nkpts,
                                             comm=serial_comm,
                                             nspins=nspins,
                                             nibzkpts=nkpts,
                                             weight_k=weight_k,
                                             collect=lambda x, broadcast: x))

    for s in range(nspins):
        for k in range(nkpts):
            kpt = SimpleNamespace(s=s,
                                  weight=weight_k[k] * 2 / nspins,
                                  eps_n=eps_skn[s, k],
                                  f_n=f_skn[s, k])
            wfs.kpt_u.append(kpt)

    occ.calculate(wfs)

    return f_skn, occ.fermilevel, occ.magmom, occ.e_entropy


def findroot(func, x, tol=1e-10):
    """Function used for locating Fermi level."""
    xmin = -np.inf
    xmax = np.inf

    # Try 10 step using the gradient:
    niter = 0
    while True:
        f, dfdx = func(x)
        if abs(f) < tol:
            return x, niter
        if f < 0.0 and x > xmin:
            xmin = x
        elif f > 0.0 and x < xmax:
            xmax = x
        dx = -f / max(dfdx, 1e-18)
        if niter == 10 or abs(dx) > 0.01 or not (xmin < x + dx < xmax):
            break  # try bisection
        x += dx
        niter += 1

    # Bracket the solution:
    if not np.isfinite(xmin):
        xmin = x
        fmin = f
        step = 0.01
        while fmin > tol:
            xmin -= step
            fmin = func(xmin)[0]
            step *= 2

    if not np.isfinite(xmax):
        xmax = x
        fmax = f
        step = 0.01
        while fmax < 0:
            xmax += step
            fmax = func(xmax)[0]
            step *= 2

    # Bisect:
    while True:
        x = (xmin + xmax) / 2
        f = func(x)[0]
        if abs(f) < tol:
            return x, niter
        if f > 0:
            xmax = x
        else:
            xmin = x
        niter += 1


class OccupationNumbers:
    """Base class for all occupation number objects."""
    def __init__(self, fixmagmom):
        self.fixmagmom = fixmagmom
        self.magmom = None  # magnetic moment
        self.e_entropy = None  # -ST
        self.e_band = None  # band energy (sum_n eps_n * f_n)
        self.fermilevel = np.nan  # Fermi level
        self.nvalence = None  # number of electrons
        self.split = 0.0  # splitting of Fermi levels from fixmagmom=True
        self.niter = 0  # number of iterations for finding Fermi level
        self.ready = False
        self.fixed_fermilevel = False

    def write(self, writer):
        writer.write(fermilevel=self.fermilevel * Hartree,
                     split=self.split * Hartree)

    def read(self, reader):
        o = reader.occupations
        self.fermilevel = o.fermilevel / reader.ha
        self.split = o.split / reader.ha

    def extrapolate_energy_to_zero_width(self, e_free):
        return e_free

    def calculate(self, wfs):
        """Calculate everything.

        The following is calculated:

        * occupation numbers
        * magnetic moment
        * entropy
        * band energy
        * Fermi level
        """

        # Allow subclasses to adjust nvalence:
        self.set_number_of_electrons(wfs)

        # Allocate:
        for kpt in wfs.kpt_u:
            if kpt.f_n is None:
                kpt.f_n = wfs.bd.empty()

            # There are no eigenvalues, might as well return
            if kpt.eps_n is None:
                return

            # Sanity check.  This class will typically be the first to
            # suffer if any NaNs sneak in.
            assert not np.isnan(kpt.eps_n).any()

        fermilevel = self.fermilevel  # save for later

        # Let the master domain do the work and broadcast results:
        data = np.empty(5)
        if wfs.gd.comm.rank == 0:
            self.calculate_occupation_numbers(wfs)
            self.calculate_band_energy(wfs)
            data[:] = [self.magmom, self.e_entropy, self.e_band,
                       self.fermilevel, self.split]
        wfs.world.broadcast(data, 0)
        (self.magmom, self.e_entropy, self.e_band,
         self.fermilevel, self.split) = data

        for kpt in wfs.kpt_u:
            wfs.gd.comm.broadcast(kpt.f_n, 0)

        if self.fixed_fermilevel:
            self.fermilevel = fermilevel

    def set_number_of_electrons(self, wfs):
        self.nvalence = wfs.nvalence
        self.ready = True

    def calculate_occupation_numbers(self, wfs):
        raise NotImplementedError

    def calculate_band_energy(self, wfs):
        """Sum up all eigenvalues weighted with occupation numbers"""
        e_band = 0.0
        for kpt in wfs.kpt_u:
            e_band += np.dot(kpt.f_n, kpt.eps_n)
        self.e_band = wfs.kptband_comm.sum(e_band)

    def get_fermi_level(self):
        raise ValueError('Can not calculate Fermi level!')


class ZeroKelvin(OccupationNumbers):
    def __init__(self, fixmagmom):
        self.width = 0.0
        OccupationNumbers.__init__(self, fixmagmom)

    def calculate_occupation_numbers(self, wfs):
        if wfs.nspins == 1:
            self.spin_paired(wfs)
        elif self.fixmagmom:
            assert wfs.kd.gamma
            self.fixed_moment(wfs)
        else:
            assert wfs.kd.nibzkpts == 1
            self.spin_polarized(wfs)

        self.e_entropy = 0.0

    def occupy(self, f_n, eps_n, ne, weight=1):
        """Fill in occupation numbers.

        return HOMO and LUMO energies."""

        N = len(f_n)
        if ne == N:
            f_n[:] = weight
            return eps_n[-1], np.inf

        n = int(ne)
        f = ne - n
        f_n[:n] = weight
        assert n < N
        f_n[n] = f * weight
        f_n[n + 1:] = 0.0
        if f > 0.0:
            return eps_n[n], eps_n[n]
        return eps_n[n - 1], eps_n[n]

    def summary(self, log):
        if np.isfinite(self.fermilevel):
            if self.fixed_fermilevel:
                log('Fixed ', end='')
            if not self.fixmagmom:
                log('Fermi level: %.5f\n' % (Hartree * self.fermilevel))
            else:
                log('Fermi levels: %.5f, %.5f\n' %
                    (Hartree * (self.fermilevel + 0.5 * self.split),
                     Hartree * (self.fermilevel - 0.5 * self.split)))

    def get_fermi_level(self):
        """This function returns the calculated fermi-level.

        Care: you get two distinct fermi-levels if you do
        fixed-magmom calculations. Therefore you should use
        "get_fermi_levels" or "get_fermi_levels_mean" in
        conjunction with "get_fermi_splitting" if you do
        fixed-magmom calculations. We will issue an warning
        otherwise.

        """
        if not np.isfinite(self.fermilevel):
            OccupationNumbers.get_fermi_level(self)  # fail
        else:
            if self.fixmagmom:
                warnings.warn('Please use get_fermi_levels when ' +
                              'using fixmagmom', DeprecationWarning)
                fermilevels = np.empty(2)
                fermilevels[0] = self.fermilevel + 0.5 * self.split
                fermilevels[1] = self.fermilevel - 0.5 * self.split
                return fermilevels
            else:
                return self.fermilevel

    def get_fermi_levels(self):
        """Getting fermi-levels in case of fixed-magmom."""
        if not np.isfinite(self.fermilevel):
            OccupationNumbers.get_fermi_level(self)  # fail
        else:
            if self.fixmagmom:
                fermilevels = np.empty(2)
                fermilevels[0] = self.fermilevel + 0.5 * self.split
                fermilevels[1] = self.fermilevel - 0.5 * self.split
                return fermilevels
            else:
                raise ValueError('Distinct fermi-levels are only vaild ' +
                                 'for fixed-magmom calculations!')

    def get_fermi_levels_mean(self):
        if not np.isfinite(self.fermilevel):
            OccupationNumbers.get_fermi_level(self)  # fail
        else:
            return self.fermilevel

    def get_fermi_splitting(self):
        """Return the splitting of the fermi level in hartree.

        Returns 0.0 if calculation is not done using
        fixmagmom.

        """
        if self.fixmagmom:
            return self.split
        else:
            return 0.0

    def fixed_moment(self, wfs):
        assert wfs.nspins == 2 and wfs.kd.nbzkpts == 1
        fermilevels = np.zeros(2)
        for kpt in wfs.kpt_u:
            eps_n = wfs.bd.collect(kpt.eps_n)
            if eps_n is None:
                f_n = None
            else:
                f_n = wfs.bd.empty(global_array=True)
                sign = 1 - kpt.s * 2
                ne = 0.5 * (self.nvalence + sign * self.magmom)

                homo, lumo = self.occupy(f_n, eps_n, ne)

                fermilevels[kpt.s] = 0.5 * (homo + lumo)
            wfs.bd.distribute(f_n, kpt.f_n)
        wfs.kptband_comm.sum(fermilevels)
        self.fermilevel = fermilevels.mean()
        self.split = fermilevels[0] - fermilevels[1]

    def spin_paired(self, wfs):
        homo = -np.inf
        lumo = np.inf
        if wfs.collinear:
            ne = 0.5 * self.nvalence
        else:
            ne = self.nvalence
        for kpt in wfs.kpt_u:
            eps_n = wfs.bd.collect(kpt.eps_n)
            if wfs.bd.comm.rank == 0:
                f_n = wfs.bd.empty(global_array=True)
                hom, lum = self.occupy(f_n, eps_n,
                                       ne, kpt.weight)
                homo = max(homo, hom)
                lumo = min(lumo, lum)
            else:
                f_n = None
                self.fermilevel = np.nan
            wfs.bd.distribute(f_n, kpt.f_n)

        if wfs.bd.comm.rank == 0:
            homo = wfs.kd.comm.max(homo)
            lumo = wfs.kd.comm.min(lumo)
            self.fermilevel = 0.5 * (homo + lumo)

        self.magmom = 0.0

    def spin_polarized(self, wfs):
        eps_un = [wfs.bd.collect(kpt.eps_n) for kpt in wfs.kpt_u]
        self.fermilevel = np.nan
        nbands = wfs.bd.nbands
        if wfs.bd.comm.rank == 0:
            if wfs.kd.comm.size == 2:
                if wfs.kd.comm.rank == 1:
                    wfs.kd.comm.send(eps_un[0], 0)
                else:
                    eps_sn = [eps_un[0], np.empty(nbands)]
                    wfs.kd.comm.receive(eps_sn[1], 1)
            else:
                eps_sn = eps_un

            if wfs.kd.comm.rank == 0:
                eps_n = np.ravel(eps_sn)
                f_n = np.empty(nbands * 2)
                nsorted = eps_n.argsort()
                homo, lumo = self.occupy(f_n, eps_n[nsorted], self.nvalence)
                f_sn = f_n[nsorted.argsort()].reshape((2, nbands))
                self.magmom = f_sn[0].sum() - f_sn[1].sum()
                self.fermilevel = 0.5 * (homo + lumo)

            if wfs.kd.comm.size == 2:
                if wfs.kd.comm.rank == 0:
                    wfs.kd.comm.send(f_sn[1], 1)
                else:
                    f_sn = [None, np.empty(nbands)]
                    wfs.kd.comm.receive(f_sn[1], 0)
        else:
            f_sn = [None, None]

        for kpt in wfs.kpt_u:
            wfs.bd.distribute(f_sn[kpt.s], kpt.f_n)


class SmoothDistribution(ZeroKelvin):
    """Base class for Fermi-Dirac and other smooth distributions."""
    def __init__(self, width, fixmagmom):
        """Smooth distribution.

        Find the Fermi level by integrating in energy until
        the number of electrons is correct.

        width: float
            Width of distribution in eV.
        fixmagmom: bool
            Fix spin moment calculations.  A separate Fermi level for
            spin up and down electrons is found: self.fermilevel +
            self.split and self.fermilevel - self.split.
        """

        ZeroKelvin.__init__(self, fixmagmom)
        self.width = width / Hartree

    def todict(self):
        dct = {'width': self.width * Hartree}
        if self.fixmagmom:
            dct['fixmagmom'] = True
        return dct

    def __str__(self):
        s = 'Occupation numbers:\n'
        if self.fixmagmom:
            s += '  Fixed magnetic moment\n'
        if self.fixed_fermilevel:
            s += '  Fixed Fermi level\n'
        return s

    def calculate_occupation_numbers(self, wfs):
        if self.width == 0 or self.nvalence == wfs.bd.nbands * 2:
            ZeroKelvin.calculate_occupation_numbers(self, wfs)
            return

        if not np.isfinite(self.fermilevel):
            self.fermilevel = self.guess_fermi_level(wfs)

        if not self.fixmagmom or wfs.nspins == 1:
            result = self.find_fermi_level(wfs, self.nvalence, self.fermilevel)
            self.fermilevel, self.magmom, self.e_entropy = result

            if wfs.nspins == 1:
                self.magmom = 0.0
        else:
            fermilevels = np.empty(2)
            self.e_entropy = 0.0
            for s in range(2):
                sign = 1 - s * 2
                ne = 0.5 * (self.nvalence + sign * self.magmom)
                fermilevel = self.fermilevel + 0.5 * sign * self.split
                fermilevels[s], magmom, e_entropy = \
                    self.find_fermi_level(wfs, ne, fermilevel, [s])
                self.e_entropy += e_entropy
            self.fermilevel = fermilevels.mean()
            self.split = fermilevels[0] - fermilevels[1]

    def guess_fermi_level(self, wfs):
        fermilevel = 0.0

        kd = wfs.kd

        myeps_un = np.empty((kd.mynks, wfs.bd.nbands))
        for u, kpt in enumerate(wfs.kpt_u):
            myeps_un[u] = wfs.bd.collect(kpt.eps_n)

        if wfs.bd.comm.rank == 0:
            eps_skn = kd.collect(myeps_un, broadcast=False)
            if kd.comm.rank == 0:
                eps_n = eps_skn.ravel()
                w_skn = np.empty((kd.nspins, kd.nibzkpts, wfs.bd.nbands))
                w_skn[:] = (2.0 / wfs.nspins * kd.weight_k[:, np.newaxis])
                w_n = w_skn.ravel()
                n_i = eps_n.argsort()
                w_i = w_n[n_i]
                f_i = np.add.accumulate(w_i) - 0.5 * w_i
                i = np.nonzero(f_i >= self.nvalence)[0][0]
                if i == 0:
                    fermilevel = eps_n[n_i[0]]
                else:
                    fermilevel = ((eps_n[n_i[i]] *
                                   (self.nvalence - f_i[i - 1]) +
                                   eps_n[n_i[i - 1]] *
                                   (f_i[i] - self.nvalence)) /
                                  (f_i[i] - f_i[i - 1]))

        # XXX broadcast would be better!
        return wfs.kptband_comm.sum(fermilevel)

    def find_fermi_level(self, wfs, ne, fermilevel, spins={0, 1, None}):
        niter = 0

        x = self.fermilevel
        if not np.isfinite(x):
            x = self.guess_fermi_level(wfs)

        data = np.empty(4)

        def f(x, data=data):
            data[:] = 0.0
            for kpt in wfs.kpt_u:
                if kpt.s in spins:
                    data += self.distribution(kpt, x)
            wfs.kptband_comm.sum(data)
            n, dnde = data[:2]
            dn = n - ne
            return dn, dnde

        fermilevel, niter = findroot(f, x)

        self.niter = niter
        magmom, e_entropy = data[2:]
        return fermilevel, magmom, e_entropy


class FermiDirac(SmoothDistribution):
    def __init__(self, width, fixmagmom=False):
        SmoothDistribution.__init__(self, width, fixmagmom)

    def todict(self):
        dct = SmoothDistribution.todict(self)
        dct['name'] = 'fermi-dirac'
        return dct

    def __str__(self):
        s = '  Fermi-Dirac: width={0:.4f} eV\n'.format(self.width * Hartree)
        return SmoothDistribution.__str__(self) + s

    def distribution(self, kpt, fermilevel):
        x = (kpt.eps_n - fermilevel) / self.width
        x = x.clip(-100, 100)
        y = np.exp(x)
        z = y + 1.0
        kpt.f_n[:] = kpt.weight / z
        n = kpt.f_n.sum()
        dnde = (n - (kpt.f_n**2).sum() / kpt.weight) / self.width
        y *= x
        y /= z
        y -= np.log(z)
        e_entropy = kpt.weight * y.sum() * self.width
        sign = 1 - kpt.s * 2 if kpt.s is not None else 0.0
        return np.array([n, dnde, n * sign, e_entropy])

    def extrapolate_energy_to_zero_width(self, E):
        return E - 0.5 * self.e_entropy


class MethfesselPaxton(SmoothDistribution):
    def __init__(self, width, order=0, fixmagmom=False):
        SmoothDistribution.__init__(self, width, fixmagmom)
        self.order = order

    def todict(self):
        dct = SmoothDistribution.todict(self)
        dct['name'] = 'methfessel-paxton'
        dct['order'] = self.order
        return dct

    def __str__(self):
        s = '  Methfessel-Paxton: width={0:.4f} eV, order={1}\n'.format(
            self.width * Hartree, self.order)
        return SmoothDistribution.__str__(self) + s

    def distribution(self, kpt, fermilevel):
        x = (kpt.eps_n - fermilevel) / self.width
        x = x.clip(-100, 100)

        z = 0.5 * (1 - erf(x))
        for i in range(self.order):
            z += (self.coff_function(i + 1) *
                  self.hermite_poly(2 * i + 1, x) * np.exp(-x**2))
        kpt.f_n[:] = kpt.weight * z
        n = kpt.f_n.sum()

        dnde = 1 / np.sqrt(pi) * np.exp(-x**2)
        for i in range(self.order):
            dnde += (self.coff_function(i + 1) *
                     self.hermite_poly(2 * i + 2, x) * np.exp(-x**2))
        dnde = dnde.sum()
        dnde *= kpt.weight / self.width
        e_entropy = (0.5 * self.coff_function(self.order) *
                     self.hermite_poly(2 * self.order, x) * np.exp(-x**2))
        e_entropy = -kpt.weight * e_entropy.sum() * self.width

        sign = 1 - kpt.s * 2
        return np.array([n, dnde, n * sign, e_entropy])

    def coff_function(self, n):
        return (-1)**n / (np.product(np.arange(1, n + 1)) *
                          4**n * np.sqrt(np.pi))

    def hermite_poly(self, n, x):
        if n == 0:
            return 1
        elif n == 1:
            return 2 * x
        else:
            return (2 * x * self.hermite_poly(n - 1, x) -
                    2 * (n - 1) * self.hermite_poly(n - 2, x))

    def extrapolate_energy_to_zero_width(self, E):
        return E - self.e_entropy / (self.order + 2)


class FixedOccupations(ZeroKelvin):
    def __init__(self, occupation):
        self.occupation = np.array(occupation)
        ZeroKelvin.__init__(self, True)

    def spin_paired(self, wfs):
        return self.fixed_moment(wfs)

    def fixed_moment(self, wfs):
        for kpt in wfs.kpt_u:
            wfs.bd.distribute(self.occupation[kpt.s], kpt.f_n)


class TFOccupations(FermiDirac):
    def __init__(self):
        FermiDirac.__init__(self, width=0.0, fixmagmom=False)

    def todict(self):
        return {'name': 'orbital-free'}

    def occupy(self, f_n, eps_n, ne, weight=1):
        """Fill in occupation numbers.

        In TF mode only one band. Is guaranteed to work only
        for spin-paired case.

        return HOMO and LUMO energies."""
        # Same as occupy in FermiDirac expect one band: weight = ne
        return FermiDirac.occupy(self, f_n, eps_n, 1, ne * weight)



import numpy as np
from ase import Atoms
from ase.units import Bohr, Hartree
from gpaw.jellium import JelliumSlab
from gpaw import GPAW, Mixer
from gpaw.test import equal

rs = 5.0 * Bohr  # Wigner-Seitz radius
h = 0.24          # grid-spacing
a = 8 * h        # lattice constant
v = 3 * a        # vacuum
L = 8 * a       # thickness
k = 6           # number of k-points (k*k*1)

ne = a**2 * L / (4 * np.pi / 3 * rs**3)

bc = JelliumSlab(ne, z1=v, z2=v + L)

surf = Atoms(pbc=(True, True, False),
             cell=(a, a, v + L + v))
surf.calc = GPAW(background_charge=bc,
                 poissonsolver={'dipolelayer': 'xy'},
                 xc='LDA_X+LDA_C_WIGNER',
                 eigensolver='dav',
                 kpts=[k, k, 1],
                 h=h,
                 maxiter=300,
                 convergence={'density': 1e-5},
                 mixer=Mixer(0.3, 7, 100),
                 nbands=int(ne / 2) + 15,
                 txt='surface.txt')
e = surf.get_potential_energy()

efermi = surf.calc.get_fermi_level()
# Get (x-y-averaged) electrostatic potential
# Must collect it from the CPUs
# https://listserv.fysik.dtu.dk/pipermail/gpaw-users/2014-January/002524.html
ham = surf.calc.hamiltonian
v = (ham.finegd.collect(ham.vHt_g,
                        broadcast=True) * Hartree).mean(0).mean(0)

# Get the work function
phi1 = v[-1] - efermi

equal(phi1, 2.7162036108, 1e-3)
# Reference value: Lang and Kohn, 1971, Theory of Metal Surfaces: Work function
# DOI 10.1103/PhysRevB.3.1215
# r_s = 5, work function = 2.73 eV

surf.calc.write('jellium.gpw')



from __future__ import division, print_function

import json
import sys
from math import pi
from pathlib import Path

import numpy as np
from ase.units import Hartree

import gpaw.mpi as mpi
from gpaw.xc import XC
from gpaw.xc.tools import vxc
from gpaw.xc.kernel import XCNull
from gpaw.response.pair import PairDensity
from gpaw.wavefunctions.pw import PWDescriptor
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.utilities import unpack, unpack2, packed_index
from gpaw.response.wstc import WignerSeitzTruncatedCoulomb


def pawexxvv(atomdata, D_ii):
    """PAW correction for valence-valence EXX energy."""
    ni = len(D_ii)
    V_ii = np.empty((ni, ni))
    for i1 in range(ni):
        for i2 in range(ni):
            V = 0.0
            for i3 in range(ni):
                p13 = packed_index(i1, i3, ni)
                for i4 in range(ni):
                    p24 = packed_index(i2, i4, ni)
                    V += atomdata.M_pp[p13, p24] * D_ii[i3, i4]
            V_ii[i1, i2] = V
    return V_ii


def select_kpts(kpts, calc):
    """Function to process input parameters that take a list of k-points given
    in different format and returns a list of indices of the corresponding
    k-points in the IBZ."""
    if kpts is None:
        # Do all k-points in the IBZ:
        return np.arange(calc.wfs.kd.nibzkpts)

    if np.asarray(kpts).ndim == 1:
        return kpts

    # Find k-points:
    bzk_Kc = calc.get_bz_k_points()
    indices = []
    for k_c in kpts:
        d_Kc = bzk_Kc - k_c
        d_Kc -= d_Kc.round()
        K = abs(d_Kc).sum(1).argmin()
        if not np.allclose(d_Kc[K], 0):
            raise ValueError('Could not find k-point: {k_c}'
                             .format(k_c=k_c))
        k = calc.wfs.kd.bz2ibz_k[K]
        indices.append(k)
    return indices


class EXX(PairDensity):
    def __init__(self, calc, xc=None, kpts=None, bands=None, ecut=None,
                 stencil=2,
                 omega=None, world=mpi.world, txt=sys.stdout, timer=None):
        """Non self-consistent hybrid functional calculations.

        Eigenvalues and total energy can be calculated.

        calc: str or PAW object
            GPAW calculator object or filename of saved calculator object.
        xc: str
            Name of functional.  Use one of EXX, PBE0, HSE03, HSE06 or B3LYP.
            Default is EXX.
        kpts: list of in or list of list of int
            List of indices of the IBZ k-points to calculate the quasi particle
            energies for.  Default is all k-points.  One can also specify the
            coordiantes of the k-point.  As an example, Gamma and X for an
            FCC crystal would be: kpts=[[0, 0, 0], [1 / 2, 1 / 2, 0]].
        bands: tuple of two ints
            Range of band indices, like (n1, n2), to calculate the quasi
            particle energies for. Bands n where n1<=n<n2 will be
            calculated.  Note that the second band index is not included.
            Default is all occupied bands.
        ecut: float
            Plane wave cut-off energy in eV.  Default it the same as was used
            for the ground-state calculations.
        """

        PairDensity.__init__(self, calc, ecut, world=world, txt=txt,
                             timer=timer)

        def _xc(name):
            return {'name': name, 'stencil': stencil}

        if xc is None or xc == 'EXX':
            self.exx_fraction = 1.0
            xc = XC(XCNull())
        elif xc == 'PBE0':
            self.exx_fraction = 0.25
            xc = XC(_xc('HYB_GGA_XC_PBEH'))
        elif xc == 'HSE03':
            omega = 0.106
            self.exx_fraction = 0.25
            xc = XC(_xc('HYB_GGA_XC_HSE03'))
        elif xc == 'HSE06':
            omega = 0.11
            self.exx_fraction = 0.25
            xc = XC(_xc('HYB_GGA_XC_HSE06'))
        elif xc == 'B3LYP':
            self.exx_fraction = 0.2
            xc = XC(_xc('HYB_GGA_XC_B3LYP'))

        self.xc = xc
        self.omega = omega
        self.exc = np.nan  # density dependent part of xc-energy

        self.kpts = select_kpts(kpts, self.calc)

        if bands is None:
            # Do all occupied bands:
            bands = [0, self.nocc2]

        print('Calculating exact exchange contributions for band index',
              '%d-%d' % (bands[0], bands[1] - 1), file=self.fd)
        print('for IBZ k-points with indices:',
              ', '.join(str(i) for i in self.kpts), file=self.fd)

        self.bands = bands

        if self.ecut is None:
            self.ecut = self.calc.wfs.pd.ecut
        print('Plane-wave cutoff: %.3f eV' % (self.ecut * Hartree),
              file=self.fd)

        shape = (self.calc.wfs.nspins, len(self.kpts), bands[1] - bands[0])
        self.exxvv_sin = np.zeros(shape)   # valence-valence exchange energies
        self.exxvc_sin = np.zeros(shape)   # valence-core exchange energies
        self.f_sin = np.empty(shape)       # occupation numbers

        # The total EXX energy will not be calculated if we are only
        # interested in a few eigenvalues for a few k-points
        self.exx = np.nan    # total EXX energy
        self.exxvv = np.nan  # valence-valence
        self.exxvc = np.nan  # valence-core
        self.exxcc = 0.0     # core-core

        self.mysKn1n2 = None  # my (s, K, n1, n2) indices
        self.distribute_k_points_and_bands(0, self.nocc2)

        # All occupied states:
        self.mykpts = [self.get_k_point(s, K, n1, n2)
                       for s, K, n1, n2 in self.mysKn1n2]

        if omega is None:
            print('Using Wigner-Seitz truncated coulomb interaction.',
                  file=self.fd)
            self.wstc = WignerSeitzTruncatedCoulomb(self.calc.wfs.gd.cell_cv,
                                                    self.calc.wfs.kd.N_c,
                                                    self.fd)
        self.iG_qG = {}  # cache

        # PAW matrices:
        self.V_asii = []  # valence-valence correction
        self.C_aii = []   # valence-core correction
        self.initialize_paw_exx_corrections()

    def calculate(self, restart: str = None) -> None:
        """Do the calculation.

        restart_filename: str or Path
            Name of restart json-file.  Allows for incremental calculation
            of eigenvalues.
        """
        kd = self.calc.wfs.kd
        nspins = self.calc.wfs.nspins

        if restart:
            if isinstance(restart, str):
                restart = Path(restart)

            if restart.is_file():
                data = json.loads(restart.read_text())
                n = len(data)
                print('Restarting from {restart}.  '
                      'Read {n} spin+k-point pairs.'
                      .format(**locals()),
                      file=self.fd)
            else:
                data = []
        else:
            data = []

        n = 0
        for s in range(nspins):
            for i, k1 in enumerate(self.kpts):
                if n < len(data):
                    f_n, exxvv_n, exxvc_n = data[n]
                    self.f_sin[s, i] = f_n
                    self.exxvc_sin[s, i] = exxvc_n
                    self.exxvv_sin[s, i] = exxvv_n
                    n += 1
                    continue

                K1 = kd.ibz2bz_k[k1]
                kpt1 = self.get_k_point(s, K1, *self.bands)
                self.f_sin[s, i] = kpt1.f_n
                for kpt2 in self.mykpts:
                    if kpt2.s == s:
                        self.calculate_q(i, kpt1, kpt2)

                self.calculate_paw_exx_corrections(i, kpt1)

                self.world.sum(self.exxvv_sin[s, i])

                if restart and self.world.rank == 0:
                    data.append([x_n.tolist()
                                 for x_n in [self.f_sin[s, i],
                                             self.exxvc_sin[s, i],
                                             self.exxvv_sin[s, i]]])
                    tmp = restart.with_name(restart.name + '.tmp')
                    tmp.write_text(json.dumps(data))
                    # Overwrite restart-file in in almost atomic step that
                    # hopefully does not crash due to job running out of time:
                    tmp.rename(restart)

                n += 1

        # Calculate total energy if we have everything needed:
        if (len(self.kpts) == kd.nibzkpts and
            self.bands[0] == 0 and
            self.bands[1] >= self.nocc2):
            exxvv_i = (self.exxvv_sin * self.f_sin).sum(axis=2).sum(axis=0)
            exxvc_i = 2 * (self.exxvc_sin * self.f_sin).sum(axis=2).sum(axis=0)
            self.exxvv = np.dot(kd.weight_k[self.kpts], exxvv_i) / nspins
            self.exxvc = np.dot(kd.weight_k[self.kpts], exxvc_i) / nspins
            self.exx = self.exxvv + self.exxvc + self.exxcc
            print('Exact exchange energy:', file=self.fd)
            for kind, exx in [('valence-valence', self.exxvv),
                              ('valence-core', self.exxvc),
                              ('core-core', self.exxcc),
                              ('total', self.exx)]:
                print('%16s%11.3f eV' % (kind + ':', exx * Hartree),
                      file=self.fd)

            self.exc = self.calculate_hybrid_correction()

        exx_sin = self.exxvv_sin + self.exxvc_sin
        print('EXX eigenvalue contributions in eV:', file=self.fd)
        print(np.array_str(exx_sin * Hartree, precision=3), file=self.fd)

    def get_exx_energy(self):
        return self.exx * Hartree

    def get_total_energy(self):
        ham = self.calc.hamiltonian
        return (self.exx * self.exx_fraction + self.exc +
                ham.e_total_free - ham.e_xc) * Hartree

    def get_eigenvalue_contributions(self):
        b1, b2 = self.bands
        e_sin = vxc(self.calc, self.xc)[:, self.kpts, b1:b2] / Hartree
        e_sin += (self.exxvv_sin + self.exxvc_sin) * self.exx_fraction
        return e_sin * Hartree

    def calculate_q(self, i, kpt1, kpt2):
        wfs = self.calc.wfs

        q_c = wfs.kd.bzk_kc[kpt2.K] - wfs.kd.bzk_kc[kpt1.K]
        qd = KPointDescriptor([q_c])
        pd = PWDescriptor(self.ecut, wfs.gd, wfs.dtype, kd=qd)

        Q_G = self.get_fft_indices(kpt1.K, kpt2.K, q_c, pd,
                                   kpt1.shift_c - kpt2.shift_c)

        Q_aGii = self.initialize_paw_corrections(pd, soft=True)

        for n in range(kpt1.n2 - kpt1.n1):
            ut1cc_R = kpt1.ut_nR[n].conj()
            C1_aGi = [np.dot(Q_Gii, P1_ni[n].conj())
                      for Q_Gii, P1_ni in zip(Q_aGii, kpt1.P_ani)]
            n_mG = self.calculate_pair_densities(ut1cc_R, C1_aGi, kpt2,
                                                 pd, Q_G)
            e = self.calculate_n(pd, n, n_mG, kpt2)
            self.exxvv_sin[kpt1.s, i, n] += e

    def calculate_n(self, pd, n, n_mG, kpt2):
        iG_G = self.get_coulomb_kernel(pd)
        x = 4 * pi / self.calc.wfs.kd.nbzkpts / pd.gd.dv**2

        e = 0.0
        for f, n_G in zip(kpt2.f_n, n_mG):
            x_G = n_G * iG_G
            e -= x * f * pd.integrate(x_G, x_G).real

        return e

    def get_coulomb_kernel(self, pd):
        if self.omega is not None:
            G2_G = pd.G2_qG[0]
            iG_G = np.empty_like(G2_G)
            if pd.kd.gamma:
                iG_G[0] = 1 / (2 * self.omega)
            else:
                iG_G[0] = ((1 - np.exp(-G2_G[0] / (4 * self.omega**2))) /
                           G2_G[0])**0.5
            iG_G[1:] = ((1 - np.exp(-G2_G[1:] / (4 * self.omega**2))) /
                        G2_G[1:])**0.5
            return iG_G

        key = tuple((pd.kd.bzk_kc[0] * self.calc.wfs.kd.N_c).round())
        iG_G = self.iG_qG.get(key)
        if iG_G is None:
            v_G = self.wstc.get_potential(pd)
            iG_G = (v_G / (4 * pi))**0.5
            self.iG_qG[key] = iG_G
        return iG_G

    def initialize_paw_exx_corrections(self):
        for a, atomdata in enumerate(self.calc.wfs.setups):
            V_sii = []
            for D_p in self.calc.density.D_asp[a]:
                D_ii = unpack2(D_p)
                V_ii = pawexxvv(atomdata, D_ii)
                V_sii.append(V_ii)
            if atomdata.X_p is None:
                C_ii = D_ii * 0.0
            else:
                C_ii = unpack(atomdata.X_p)
            self.V_asii.append(V_sii)
            self.C_aii.append(C_ii)
            self.exxcc += atomdata.ExxC or 0.0

    def calculate_paw_exx_corrections(self, i, kpt):
        x = self.calc.wfs.nspins / self.world.size
        s = kpt.s

        for V_sii, C_ii, P_ni in zip(self.V_asii, self.C_aii, kpt.P_ani):
            V_ii = V_sii[s]
            v_n = (np.dot(P_ni, V_ii) * P_ni.conj()).sum(axis=1).real
            c_n = (np.dot(P_ni, C_ii) * P_ni.conj()).sum(axis=1).real
            self.exxvv_sin[s, i] -= v_n * x
            self.exxvc_sin[s, i] -= c_n

    def calculate_hybrid_correction(self):
        dens = self.calc.density
        if dens.nt_sg is None:
            dens.interpolate_pseudo_density()
        exc = self.xc.calculate(dens.finegd, dens.nt_sg)
        for a, D_sp in dens.D_asp.items():
            atomdata = dens.setups[a]
            exc += self.xc.calculate_paw_correction(atomdata, D_sp)
        return exc



# General imports
import os
import numbers
import numpy as np
# ASE and GPAW imports
from ase.units import Bohr, Hartree
from gpaw.jellium import Jellium, JelliumSlab
from gpaw.hamiltonian import RealSpaceHamiltonian
from gpaw.fd_operators import Gradient
from gpaw.dipole_correction import DipoleCorrection

# Implicit solvent imports
from gpaw.solvation.cavity import (Potential, Power12Potential,
                                   get_pbc_positions)
from gpaw.solvation.calculator import SolvationGPAW
from gpaw.solvation.hamiltonian import SolvationRealSpaceHamiltonian
from gpaw.solvation.poisson import WeightedFDPoissonSolver


class SJM(SolvationGPAW):
    """Subclass of the SolvationGPAW class which includes the
    Solvated Jellium method.

    The method allows the simulation of an electrochemical environment
    by calculating constant potential quantities on the basis of constant
    charge DFT runs. For this purpose, it allows the usage of non-neutral
    periodic slab systems. Cell neutrality is achieved by adding a
    background charge in the solvent region above the slab

    Further detail are given in http://dx.doi.org/10.1021/acs.jpcc.8b02465

    Parameters:

    ne: float
        Number of electrons added in the atomic system and (with opposite
        sign) in the background charge region. At the start it can be an
        initial guess for the needed number of electrons and will be
        changed to the current number in the course of the calculation
    potential: float
        The potential that should be reached or kept in the course of the
        calculation. If set to "None" (default) a constant charge charge
        calculation based on the value of `ne` is performed.
    potential_equilibration_mode: str
        The mode for potential relaxation (only relevant for structure
        optimizations):

        'seq' (Default):
            Sequential mode.
            In a geometry optimization, the
            potential will be fully equilibrated after each
            ionic step, if outside of dpot.

        'sim':
            Simultaneous mode.
            After each ionic step only ne is adjusted and the geometry
            optimization continues. If potential deviation from target
            is higher than `max_pot_deviation` one complete potential
            equilibration at constant geometry is triggered.

    dpot: float
        Tolerance for the deviation of the target potential in sequential
        mode. If the potential is outside the defined range `ne` will be
        changed in order to get inside again. Default: 0.01V.
    max_pot_deviation: float
        Maximum tolerance for the deviation in the potential from target
        potential in simultaneous mode. If the tolerance is surpassed a
        complete potential equilibration is triggered. Default: 0.2 V.
    doublelayer: dict
        Parameters regarding the shape of the counter charge region
        Implemented keys:

        'start': float or 'cavity_like'
            If a float is given it corresponds to the lower
            boundary coordinate (default: z), where the counter charge
            starts. If 'cavity_like' is given the counter charge will
            take the form of the cavity up to the 'upper_limit'.
        'thickness': float
            Thickness of the counter charge region in Angstrom.
            Can only be used if start is not 'cavity_like' and will
            be overwritten by 'upper_limit'.
        'upper_limit': float
            Upper boundary of the counter charge region in terms of
            coordinate in Angstrom (default: z). The default is
            atoms.cell[2][2] - 5.
    verbose: bool or 'cube'
        True:
            Write final electrostatic potential, background charge and
            and cavity into ASCII files.
        'cube':
            In addition to 'True', also write the cavity on the
            3D-grid into a cube file.
    write_grandcanonical_energy: bool
        Write the constant potential energy into output files such as
        trajectory files. Default: True

    """
    implemented_properties = ['energy', 'forces', 'stress', 'dipole',
                              'magmom', 'magmoms', 'ne', 'electrode_potential']

    def __init__(self, ne=0, doublelayer=None, potential=None,
                 write_grandcanonical_energy=True, dpot=0.01,
                 potential_equilibration_mode='seq', tiny=1e-8,
                 max_pot_deviation=0.2, verbose=False, **gpaw_kwargs):

        SolvationGPAW.__init__(self, **gpaw_kwargs)

        self.tiny = tiny
        if abs(ne) < self.tiny:
            self.ne = self.tiny
        else:
            self.ne = ne

        self.potential = potential
        self.dpot = dpot
        self.dl = doublelayer
        self.verbose = verbose
        self.previous_ne = 0
        self.write_grandcanonical = write_grandcanonical_energy
        self.previous_pot = None
        self.slope = None
        self.equil_iters = 10

        if potential_equilibration_mode in ['seq', 'sim']:
            self.equil_mode = potential_equilibration_mode
        else:
            raise ValueError("Potential equilibration mode not recognized."
                             "Implemented modes are 'seq' and 'sim'")

        self.pot_tol = max_pot_deviation

        self.log('-----------\nGPAW SJM module in %s\n----------\n'
                 % (os.path.abspath(__file__)))

    def create_hamiltonian(self, realspace, mode, xc):
        if not realspace:
            raise NotImplementedError(
                'SJM does not support '
                'calculations in reciprocal space yet.')

        dens = self.density

        self.hamiltonian = SJM_RealSpaceHamiltonian(
            *self.stuff_for_hamiltonian,
            gd=dens.gd, finegd=dens.finegd,
            nspins=dens.nspins,
            collinear=dens.collinear,
            setups=dens.setups,
            timer=self.timer,
            xc=xc,
            world=self.world,
            redistributor=dens.redistributor,
            vext=self.parameters.external,
            psolver=self.parameters.poissonsolver,
            stencil=mode.interpolation)

        self.log(self.hamiltonian)

        xc.set_grid_descriptor(self.hamiltonian.finegd)

    def set_atoms(self, atoms):
        self.atoms = atoms
        self.spos_ac = atoms.get_scaled_positions() % 1.0

    def set(self, **kwargs):
        """Change parameters for calculator.

        It differs from the standard `set` function in two ways:
        - SJM specific keywords can be set
        - It does not reinitialize and delete `self.wfs` if the
          background charge is changed.

        """

        SJM_keys = ['background_charge', 'ne', 'potential', 'dpot',
                    'doublelayer', 'potential_equilibration_mode',
                    'max_pot_deviation']

        SJM_changes = {}
        for key in kwargs:
            if key in SJM_keys:
                SJM_changes[key] = None

        for key in SJM_changes:
            SJM_changes[key] = kwargs.pop(key)

        major_changes = False
        if kwargs:
            SolvationGPAW.set(self, **kwargs)
            major_changes = True

        if any(key in SJM_keys[1:] for key in SJM_changes):
            if self.equil_mode == 'sim':
                self.log("Preequilibrating the potential",
                         "after manual change of",
                         list(SJM_changes.keys()))
                self.equil_iters = 10

        # SJM custom `set` for the new keywords
        for key in SJM_changes:

            if key in ['potential_equilibration_mode']:
                self.equil_mode = SJM_changes[key]
            if key in ['max_pot_deviation']:
                self.pot_tol = SJM_changes[key]

            if key in ['potential', 'doublelayer']:
                self.results = {}
                if key == 'potential':
                    self.potential = SJM_changes[key]
                    if self.potential is not None:
                        if self.wfs is None:
                            self.log('Target electrode potential: %1.4f V'
                                     % self.potential)
                        else:
                            self.log('New Target electrode potential: %1.4f V'
                                     % self.potential)
                    else:
                        self.log('Potential equilibration has been '
                                 'turned off')

                if key == 'doublelayer':
                    self.dl = SJM_changes[key]
                    self.set(background_charge=self.define_jellium(self.atoms))

            if key in ['dpot']:
                self.log('Potential tolerance has been changed to %1.4f V'
                         % SJM_changes[key])
                try:
                    potint = self.get_electrode_potential()
                except AttributeError:
                    pass
                else:
                    if abs(potint - self.potential) > SJM_changes[key]:
                        self.results = {}
                        self.log('Recalculating...\n')
                    else:
                        self.log('Potential already reached the criterion.\n')
                self.dpot = SJM_changes[key]

            if key in ['background_charge']:
                self.parameters[key] = SJM_changes[key]
                if self.wfs is not None:
                    if major_changes:
                        self.density = None
                    else:
                        self.density.reset()

                        self.density.background_charge = \
                            SJM_changes['background_charge']
                        self.density.background_charge.set_grid_descriptor(
                            self.density.finegd)

                        self.spos_ac = self.atoms.get_scaled_positions() % 1.0

                        self.density.mixer.reset()

                        self.wfs.initialize(self.density, self.hamiltonian,
                                            self.spos_ac)

                        self.wfs.eigensolver.reset()
                        self.scf.reset()

                        self.log('\n------------')
                        self.log('Jellium properties changed!')

                    if abs(self.ne) < self.tiny:
                        self.ne = self.tiny
                    self.wfs.nvalence += self.ne - self.previous_ne

                self.log('Current number of Excess Electrons: %1.5f' % self.ne)
                self.log('Jellium size parameters:')
                self.log('  Lower boundary: %s' % self.dl['start'])
                self.log('  Upper boundary: %s' % self.dl['upper_limit'])
                self.log('------------\n')

            if key in ['ne']:
                self.results = {}
                if abs(SJM_changes['ne']) < self.tiny:
                    self.ne = self.tiny
                    self.log('Number of Excess electrons manually changed ',
                             'to 0')
                else:
                    self.ne = SJM_changes['ne']
                    self.log('Number of Excess electrons manually changed ',
                             'to %1.8f' % (self.ne))

    def calculate(self, atoms=None, properties=['energy'],
                  system_changes=['cell'], ):
        """
        Perform a calculation with SJM

        This module includes the potential equilibration loop characteristic
        for SJM. It is essentially a wrapper around GPAW.calculate()

        """

        if self.potential:
            equil_iter = 0
            while equil_iter < self.equil_iters:
                if equil_iter == 1:
                    self.timer.start('Potential equilibration loop')
                    system_changes = []

                if self.atoms is None:
                    self.atoms = atoms

                if self.equil_iters > 1 or 'positions' in system_changes:
                    self.set_jellium(atoms)

                SolvationGPAW.calculate(self, atoms, ['energy'],
                                        system_changes)

                if self.verbose:
                    self.write_cavity_and_bckcharge()

                pot_int = self.get_electrode_potential()

                if self.equil_iters > 1:
                    if abs(pot_int - self.potential) < self.dpot:
                        self.previous_pot = pot_int
                        self.previous_ne = self.ne

                        if self.equil_mode == 'sim':
                            if self.equil_iters > 1:
                                self.log("Changing to simultaneous potential "
                                         "and geometry equilibration\n")
                                self.equil_iters = 1

                        if equil_iter > 0:
                            self.timer.stop('Potential equilibration loop')

                        if self.slope is not None:
                            self.log("Slope of potential versus ne "
                                     "(unchanged): %1.8f V/e\n"
                                     % (self.slope))

                        break

                if self.equil_iters > 1 or 'positions' in system_changes:
                    self.change_ne(pot_int)

                if abs(pot_int - self.potential) > self.pot_tol:
                    self.log('Deviation from target potential outside'
                             'of threshold, equilibrating potential')
                    self.equil_iters = 10

                equil_iter += 1

            if self.equil_iters > 1:
                if abs(pot_int - self.potential) > self.dpot:
                    raise Exception(
                        'Potential could not be reached after ten iterations. '
                        'Aborting!')

        else:
            self.set_jellium(atoms)
            SolvationGPAW.calculate(self, atoms, ['energy'],
                                    system_changes)
            if self.verbose:
                self.write_cavity_and_bckcharge()
            self.previous_ne = self.ne

        if properties != ['energy']:
            SolvationGPAW.calculate(self, atoms, properties, [])

    def change_ne(self, pot_int):
        if self.previous_pot is not None:
            if abs(self.previous_ne - self.ne) > 2 * self.tiny:
                self.slope = (pot_int - self.previous_pot) / \
                    (self.ne - self.previous_ne)
                self.log("Slope of potential versus ne: "
                         "%1.8f V/e\n" % (self.slope))

        if self.slope is None:
            self.previous_ne = self.ne
            self.ne += np.sign(pot_int - self.potential) * self.dpot
        else:
            d = (pot_int - self.potential) - (self.slope * self.ne)
            self.previous_ne = self.ne
            self.ne = - d / self.slope

        self.previous_pot = pot_int

    def write_cavity_and_bckcharge(self):
        self.write_parallel_func_in_z(self.hamiltonian.cavity.g_g,
                                      name='cavity_')
        if self.verbose == 'cube':
            self.write_parallel_func_on_grid(self.hamiltonian.cavity.g_g,
                                             atoms=self.atoms,
                                             name='cavity')

        if abs(self.ne) > self.tiny:
            self.write_parallel_func_in_z(
                self.density.background_charge.mask_g,
                name='background_charge_')

    def summary(self):
        omega_extra = Hartree * self.hamiltonian.e_el_extrapolated + \
            self.get_electrode_potential() * self.ne
        omega_free = Hartree * self.hamiltonian.e_el_free + \
            self.get_electrode_potential() * self.ne

        if self.write_grandcanonical:
            self.results['energy'] = omega_extra
            self.results['free_energy'] = omega_free
        else:
            self.results['energy'] = Hartree * \
                self.hamiltonian.e_el_extrapolated
            self.results['free_energy'] = Hartree * self.hamiltonian.e_el_free

        self.results['ne'] = self.ne
        self.results['electrode_potential'] = self.get_electrode_potential()
        self.hamiltonian.summary(self.occupations.fermilevel, self.log)

        self.log('----------------------------------------------------------')
        self.log("Grand Potential Energy (E_tot + E_solv - mu*ne):")
        self.log('Extrpol:    %s' % omega_extra)
        self.log('Free:    %s' % omega_free)
        self.log('-----------------------------------------------------------')
        if self.write_grandcanonical:
            self.log("Grand canonical energy will be written into results\n")
        else:
            self.log("Canonical energy will be written to results\n")

        self.density.summary(self.atoms, self.occupations.magmom, self.log)
        self.occupations.summary(self.log)
        self.wfs.summary(self.log)
        self.log.fd.flush()
        if self.verbose:
            self.write_parallel_func_in_z(self.hamiltonian.vHt_g * Hartree -
                                          self.get_fermi_level(),
                                          'elstat_potential_')

    def define_jellium(self, atoms):
        """Module for the definition of the explicit and counter charge

        """
        if self.dl is None:
            self.dl = {}

        if 'start' in self.dl:
            if self.dl['start'] == 'cavity_like':
                pass
            elif isinstance(self.dl['start'], numbers.Real):
                pass
            else:
                raise RuntimeError("The starting z value of the counter charge"
                                   "has to be either a number (coordinate),"
                                   "cavity_like' or not given (default: "
                                   " max(position)+3)")
        else:
            self.dl['start'] = max(atoms.positions[:, 2]) + 3.

        if 'upper_limit' in self.dl:
            pass
        elif 'thickness' in self.dl:
            if self.dl['start'] == 'cavity_like':
                raise RuntimeError("With a cavity-like counter charge only"
                                   "the keyword upper_limit(not thickness)"
                                   "can be used.")
            else:
                self.dl['upper_limit'] = self.dl['start'] + \
                    self.dl['thickness']
        else:
            self.dl['upper_limit'] = self.atoms.cell[2][2] - 5.0

        if self.dl['start'] == 'cavity_like':

            # XXX This part can definitely be improved
            if self.hamiltonian is None:
                filename = self.log.fd
                self.log.fd = None
                self.initialize(atoms)
                self.set_positions(atoms)
                self.log.fd = filename
                g_g = self.hamiltonian.cavity.g_g.copy()
                self.wfs = None
                self.density = None
                self.hamiltonian = None
                self.initialized = False
                return CavityShapedJellium(self.ne, g_g=g_g,
                                           z2=self.dl['upper_limit'])

            else:
                filename = self.log.fd
                self.log.fd = None
                self.set_positions(atoms)
                self.log.fd = filename
                return CavityShapedJellium(self.ne,
                                           g_g=self.hamiltonian.cavity.g_g,
                                           z2=self.dl['upper_limit'])

        elif isinstance(self.dl['start'], numbers.Real):
            return JelliumSlab(self.ne, z1=self.dl['start'],
                               z2=self.dl['upper_limit'])

    def get_electrode_potential(self):
        ham = self.hamiltonian
        fermilevel = self.occupations.fermilevel
        try:
            correction = ham.poisson.correction
        except AttributeError:
            wf2 = -fermilevel
        else:
            wf2 = (-fermilevel - correction) * Hartree

        return wf2  # refpot-E_f

    def set_jellium(self, atoms):
        if abs(self.previous_ne - self.ne) > 2 * self.tiny:
            if abs(self.ne) < self.tiny:
                self.ne = self.tiny
            self.set(background_charge=self.define_jellium(atoms))

    """Various tools for writing global functions"""

    def write_parallel_func_in_z(self, g, name='g_z.out'):
        gd = self.density.finegd
        from gpaw.mpi import world
        G = gd.collect(g, broadcast=False)
        if world.rank == 0:
            G_z = G.mean(0).mean(0)
            out = open(name + self.log.fd.name.split('.')[0] + '.out', 'w')
            for i, val in enumerate(G_z):
                out.writelines('%f  %1.8f\n' % ((i + 1) * gd.h_cv[2][2] * Bohr,
                               val))
            out.close()

    def write_parallel_func_on_grid(self, g, atoms=None, name='func.cube',
                                    outstyle='cube'):
        from ase.io import write
        gd = self.density.finegd
        G = gd.collect(g, broadcast=False)
        if outstyle == 'cube':
            write(name + '.cube', atoms, data=G)
        elif outstyle == 'pckl':
            import pickle
            out = open(name, 'wb')
            pickle.dump(G, out)
            out.close()


class SJMPower12Potential(Power12Potential):
    """Inverse power law potential.

    An 1 / r ** 12 repulsive potential
    taking the value u0 at the atomic radius.

    See also
    A. Held and M. Walter, J. Chem. Phys. 141, 174108 (2014).
    """
    depends_on_el_density = False
    depends_on_atomic_positions = True

    def __init__(self, atomic_radii, u0, pbc_cutoff=1e-6, tiny=1e-10,
                 H2O_layer=False, unsolv_backside=True):
        """Constructor for the SJMPower12Potential class.
        In SJM one also has the option of removing the solvent from the
        electrode backside and adding ghost plane/atoms to remove the solvent
        from the electrode-water interface.

        Parameters
        ----------
        atomic_radii : float
            Callable mapping an ase.Atoms object to an iterable of atomic radii
            in Angstroms.
        u0 : float
            Strength of the potential at the atomic radius in eV.
        pbc_cutoff : float
            Cutoff in eV for including neighbor cells in a calculation with
            periodic boundary conditions.
        H2O_layer: bool,int or 'plane' (default: False)
            True: Exclude the implicit solvent from the interface region
                between electrode and water. Ghost atoms will be added below
                the water layer.
            int: Explicitly account for the given number of water molecules
                above electrode. This is handy if H2O is directly adsorbed
                and a water layer is present in the unit cell at the same time.
            'plane': Use a plane instead of ghost atoms for freeing the
                surface.
        unsolv_backside: bool
            Exclude implicit solvent from the region behind the electrode

        """
        Potential.__init__(self)
        self.atomic_radii = atomic_radii
        self.u0 = float(u0)
        self.pbc_cutoff = float(pbc_cutoff)
        self.tiny = float(tiny)
        self.r12_a = None
        self.r_vg = None
        self.pos_aav = None
        self.del_u_del_r_vg = None
        self.atomic_radii_output = None
        self.H2O_layer = H2O_layer
        self.unsolv_backside = unsolv_backside

    def update(self, atoms, density):
        if atoms is None:
            return False
        self.r12_a = (self.atomic_radii_output / Bohr) ** 12
        r_cutoff = (self.r12_a.max() * self.u0 / self.pbc_cutoff) ** (1. / 12.)
        self.pos_aav = get_pbc_positions(atoms, r_cutoff)
        self.u_g.fill(.0)
        self.grad_u_vg.fill(.0)
        na = np.newaxis

        if self.unsolv_backside:
            # Removing solvent from electrode backside
            for z in range(self.u_g.shape[2]):
                if (self.r_vg[2, 0, 0, z] - atoms.positions[:, 2].min() /
                        Bohr < 0):
                    self.u_g[:, :, z] = np.inf
                    self.grad_u_vg[:, :, :, z] = 0

        if self.H2O_layer:
            # Add ghost coordinates and indices to pos_aav dictionary if
            # a water layer is present

            all_oxygen_ind = [atom.index for atom in atoms
                              if atom.symbol == 'O']

            # Disregard oxygens that don't belong to the water layer
            allwater_oxygen_ind = []
            for ox in all_oxygen_ind:
                nH = 0

                for i, atm in enumerate(atoms):
                    for period_atm in self.pos_aav[i]:
                        dist = period_atm * Bohr - atoms[ox].position
                        if np.linalg.norm(dist) < 1.3 and atm.symbol == 'H':
                            nH += 1

                if nH >= 2:
                    allwater_oxygen_ind.append(ox)

            # If the number of waters in the water layer is given as an input
            # (H2O_layer=i) then only the uppermost i water molecules are
            # regarded for unsolvating the interface (this is relevant if
            # water is adsorbed on the surface)
            if not isinstance(self.H2O_layer, (bool, str)):
                if self.H2O_layer % 1 < self.tiny:
                    self.H2O_layer = int(self.H2O_layer)
                else:
                    raise AttributeError('Only an integer number of water'
                                         'molecules is possible in the water'
                                         'layer')

                allwaters = atoms[allwater_oxygen_ind]
                indizes_water_ox_ind = np.argsort(allwaters.positions[:, 2],
                                                  axis=0)

                water_oxygen_ind = []
                for i in range(self.H2O_layer):
                    water_oxygen_ind.append(
                        allwater_oxygen_ind[indizes_water_ox_ind[-1 - i]])

            else:
                water_oxygen_ind = allwater_oxygen_ind

            oxygen = self.pos_aav[water_oxygen_ind[0]] * Bohr
            if len(water_oxygen_ind) > 1:
                for windex in water_oxygen_ind[1:]:
                    oxygen = np.concatenate(
                        (oxygen, self.pos_aav[windex] * Bohr))

            O_layer = []
            if self.H2O_layer == 'plane':
                # Add a virtual plane
                # XXX:The value -1.5, being the amount of vdW radii of O in
                # distance of the plane relative to the oxygens in the water
                # layer, is an empirical one and should perhaps be
                # interchangable.
                # For some reason the poissonsolver has trouble converging
                # sometimes if this scheme is used

                plane_rel_oxygen = -1.5 * self.atomic_radii_output[
                    water_oxygen_ind[0]]
                plane_z = oxygen[:, 2].min() + plane_rel_oxygen

                r_diff_zg = self.r_vg[2, :, :, :] - plane_z / Bohr
                r_diff_zg[r_diff_zg < self.tiny] = self.tiny
                r_diff_zg = r_diff_zg ** 2
                u_g = self.r12_a[water_oxygen_ind[0]] / r_diff_zg ** 6
                self.u_g += u_g
                u_g /= r_diff_zg
                r_diff_zg *= u_g
                self.grad_u_vg[2, :, :, :] += r_diff_zg

            else:
                # Ghost atoms are added below the explicit water layer
                cell = atoms.cell.copy() / Bohr
                cell[2][2] = 1.
                natoms_in_plane = [round(np.linalg.norm(cell[0]) * 1.5),
                                   round(np.linalg.norm(cell[1]) * 1.5)]

                plane_z = (oxygen[:, 2].min() - 1.75 *
                           self.atomic_radii_output[water_oxygen_ind[0]])
                nghatoms_z = int(round(oxygen[:, 2].min() -
                                 atoms.positions[:, 2].min()))

                for i in range(int(natoms_in_plane[0])):
                    for j in range(int(natoms_in_plane[1])):
                        for k in np.linspace(atoms.positions[:, 2].min(),
                                             plane_z, num=nghatoms_z):

                            O_layer.append(np.dot(np.array(
                                [(1.5 * i - natoms_in_plane[0] / 4) /
                                 natoms_in_plane[0],
                                 (1.5 * j - natoms_in_plane[1] / 4) /
                                 natoms_in_plane[1],
                                 k / Bohr]), cell))

            # Add additional ghost O-atoms below the actual water O atoms
            # of water which frees the interface in case of corrugated
            # water layers
            for ox in oxygen / Bohr:
                O_layer.append([ox[0], ox[1], ox[2] - 1.0 *
                                self.atomic_radii_output[
                                    water_oxygen_ind[0]] / Bohr])

            r12_add = []
            for i in range(len(O_layer)):
                self.pos_aav[len(atoms) + i] = [O_layer[i]]
                r12_add.append(self.r12_a[water_oxygen_ind[0]])
            r12_add = np.array(r12_add)
            # r12_a must have same dimensions as pos_aav items
            self.r12_a = np.concatenate((self.r12_a, r12_add))

        for index, pos_av in self.pos_aav.items():
            pos_av = np.array(pos_av)
            r12 = self.r12_a[index]
            for pos_v in pos_av:
                origin_vg = pos_v[:, na, na, na]
                r_diff_vg = self.r_vg - origin_vg
                r_diff2_g = (r_diff_vg ** 2).sum(0)
                r_diff2_g[r_diff2_g < self.tiny] = self.tiny
                u_g = r12 / r_diff2_g ** 6
                self.u_g += u_g
                u_g /= r_diff2_g
                r_diff_vg *= u_g[na, ...]
                self.grad_u_vg += r_diff_vg

        self.u_g *= self.u0 / Hartree
        self.grad_u_vg *= -12. * self.u0 / Hartree
        self.grad_u_vg[self.grad_u_vg < -1e20] = -1e20
        self.grad_u_vg[self.grad_u_vg > 1e20] = 1e20

        return True


class SJM_RealSpaceHamiltonian(SolvationRealSpaceHamiltonian):
    """Realspace Hamiltonian with continuum solvent model in the context of SJM.

    See also Section III of
    A. Held and M. Walter, J. Chem. Phys. 141, 174108 (2014).

    In contrast to the standard implicit solvent model a dipole correction can
    also be applied.

    """

    def __init__(self, cavity, dielectric, interactions, gd, finegd, nspins,
                 setups, timer, xc, world, redistributor, vext=None,
                 psolver=None, stencil=3, collinear=None):
        """Constructor of SJM_RealSpaceHamiltonian class.


        Notes
        -----
        The only difference to SolvationRealSpaceHamiltonian is the
        possibility to perform a dipole correction

        """

        self.cavity = cavity
        self.dielectric = dielectric
        self.interactions = interactions
        cavity.set_grid_descriptor(finegd)
        dielectric.set_grid_descriptor(finegd)
        for ia in interactions:
            ia.set_grid_descriptor(finegd)

        if psolver is None:
            psolver = WeightedFDPoissonSolver()
            self.dipcorr = False
        elif isinstance(psolver, dict):
            psolver = SJMDipoleCorrection(WeightedFDPoissonSolver(),
                                          psolver['dipolelayer'])
            self.dipcorr = True

        if self.dipcorr:
            psolver.poissonsolver.set_dielectric(self.dielectric)
        else:
            psolver.set_dielectric(self.dielectric)

        self.gradient = None

        RealSpaceHamiltonian.__init__(
            self,
            gd, finegd, nspins, collinear, setups, timer, xc, world,
            vext=vext, psolver=psolver,
            stencil=stencil, redistributor=redistributor)

        for ia in interactions:
            setattr(self, 'e_' + ia.subscript, None)
        self.new_atoms = None
        self.vt_ia_g = None
        self.e_el_free = None
        self.e_el_extrapolated = None

    def initialize(self):
        if self.dipcorr:
            self.gradient = [Gradient(self.finegd, i, 1.0,
                             self.poisson.poissonsolver.nn)
                             for i in (0, 1, 2)]
        else:
            self.gradient = [Gradient(self.finegd, i, 1.0,
                             self.poisson.nn)
                             for i in (0, 1, 2)]

        self.vt_ia_g = self.finegd.zeros()
        self.cavity.allocate()
        self.dielectric.allocate()
        for ia in self.interactions:
            ia.allocate()
        RealSpaceHamiltonian.initialize(self)


"""Changed module which makes it possible to use the cavity shaped
   counter charge
"""


def create_background_charge(**kwargs):
    if 'z1' in kwargs:
        return JelliumSlab(**kwargs)
    elif 'g_g' in kwargs:
        return CavityShapedJellium(**kwargs)
    return Jellium(**kwargs)


class CavityShapedJellium(Jellium):
    """The Solvated Jellium object, where the counter charge takes the form
       of the cavity.
    """
    def __init__(self, charge, g_g, z2):
        """Put the positive background charge where the solvent is present and
           z < z2.

        Parameters:
        ----------

        g_g: array
            The g function from the implicit solvent model, representing the
            percentage of the actual dielectric constant on the grid.
        z2: float
            Position of upper surface in Angstrom units."""

        Jellium.__init__(self, charge)
        self.g_g = g_g
        self.z2 = (z2 - 0.0001) / Bohr

    def todict(self):
        dct = Jellium.todict(self)
        dct.update(z2=self.z2 * Bohr + 0.0001)
        return dct

    def get_mask(self):
        r_gv = self.gd.get_grid_point_coordinates().transpose((1, 2, 3, 0))
        mask = np.logical_not(r_gv[:, :, :, 2] > self.z2).astype(float)
        mask *= self.g_g
        return mask


class SJMDipoleCorrection(DipoleCorrection):
    """Dipole-correcting wrapper around another PoissonSolver specific for SJM.

    Iterative dipole correction class as applied in SJM.

    Notes
    -----

    The modules can easily be incorporated in the trunk version of GPAW
    by just adding the `fd_solv_solve`  and adapting the `solve` modules
    in the `DipoleCorrection` class.

    This module is currently calculating the correcting dipole potential
    iteratively and we would be very grateful if anybody could
    provide an analytical solution.

    New Parameters
    ---------
    corrterm: float
    Correction factor for the added countering dipole. This is calculated
    iteratively.

    last_corrterm: float
    Corrterm in the last iteration for getting the change of slope with change
    corrterm

    last_slope: float
    Same as for `last_corrterm`

    """
    def __init__(self, poissonsolver, direction, width=1.0):
        """Construct dipole correction object."""

        DipoleCorrection.__init__(self, poissonsolver, direction, width=1.0)
        self.corrterm = 1
        self.elcorr = None
        self.last_corrterm = None

    def solve(self, pot, dens, **kwargs):
        if isinstance(dens, np.ndarray):
            # finite-diference Poisson solver:
            if hasattr(self.poissonsolver, 'dielectric'):
                return self.fd_solv_solve(pot, dens, **kwargs)
            else:
                return self.fdsolve(pot, dens, **kwargs)
        # Plane-wave solver:
        self.pwsolve(pot, dens)

    def fd_solv_solve(self, vHt_g, rhot_g, **kwargs):

        gd = self.poissonsolver.gd
        slope_lim = 1e-8
        slope = slope_lim * 10

        dipmom = gd.calculate_dipole_moment(rhot_g)[2]

        if self.elcorr is not None:
            vHt_g[:, :] -= self.elcorr

        iters2 = self.poissonsolver.solve(vHt_g, rhot_g, **kwargs)

        sawtooth_z = self.sjm_sawtooth()
        L = gd.cell_cv[2, 2]

        while abs(slope) > slope_lim:
            vHt_g2 = vHt_g.copy()
            self.correction = 2 * np.pi * dipmom * L / \
                gd.volume * self.corrterm
            elcorr = -2 * self.correction

            elcorr *= sawtooth_z
            elcorr2 = elcorr[gd.beg_c[2]:gd.end_c[2]]
            vHt_g2[:, :] += elcorr2

            VHt_g = gd.collect(vHt_g2, broadcast=True)
            VHt_z = VHt_g.mean(0).mean(0)
            slope = VHt_z[2] - VHt_z[10]

            if abs(slope) > slope_lim:
                if self.last_corrterm is not None:
                    ds = (slope - self.last_slope) / \
                        (self.corrterm - self.last_corrterm)
                    con = slope - (ds * self.corrterm)
                    self.last_corrterm = self.corrterm
                    self.corrterm = -con / ds
                else:
                    self.last_corrterm = self.corrterm
                    self.corrterm -= slope * 10.
                self.last_slope = slope
            else:
                vHt_g[:, :] += elcorr2
                self.elcorr = elcorr2

        return iters2

    def sjm_sawtooth(self):
        gd = self.poissonsolver.gd
        c = self.c
        L = gd.cell_cv[c, c]
        step = gd.h_cv[c, c] / L

        eps_g = gd.collect(self.poissonsolver.dielectric.eps_gradeps[0],
                           broadcast=True)
        eps_z = eps_g.mean(0).mean(0)

        saw = [-0.5]
        for i, eps in enumerate(eps_z):
            saw.append(saw[i] + step / eps)
        saw = np.array(saw)
        saw /= saw[-1] + step / eps_z[-1] - saw[0]
        saw -= (saw[0] + saw[-1] + step / eps_z[-1]) / 2.
        return saw



from ase import Atoms
from gpaw import GPAW
from gpaw.fdtd.poisson_fdtd import FDTDPoissonSolver
from gpaw.fdtd.polarizable_material import PermittivityPlus, PolarizableMaterial, PolarizableSphere
from gpaw.mpi import world
from gpaw.tddft import TDDFT
from gpaw.test import equal
import numpy as np

# Accuracy
energy_eps = 0.0005

# Whole simulation cell (Angstroms)
large_cell = [20, 20, 30]

# Quantum subsystem
atom_center = np.array([10.0, 10.0, 20.0])
atoms = Atoms('Na2', [atom_center + [0.0, 0.0, -1.50],
                      atom_center + [0.0, 0.0, +1.50]])

# Permittivity file
if world.rank == 0:
    fo = open('ed.txt', 'w')
    fo.writelines(['1.20 0.20 25.0'])
    fo.close()
world.barrier()

# Classical subsystem
classical_material = PolarizableMaterial()
sphere_center = np.array([10.0, 10.0, 10.0])
classical_material.add_component(PolarizableSphere(permittivity = PermittivityPlus('ed.txt'),
                                                   center = sphere_center,
                                                   radius = 5.0
                                                   ))

# Combined Poisson solver
poissonsolver = FDTDPoissonSolver(classical_material  = classical_material,
                                  qm_spacing          = 0.40,
                                  cl_spacing          = 0.40*4,
                                  cell                = large_cell,
                                  remove_moments      = (1, 4),
                                  communicator        = world,
                                  potential_coupler   = 'Refiner')
poissonsolver.set_calculation_mode('iterate')

# Combined system
atoms.set_cell(large_cell)
atoms, qm_spacing, gpts = poissonsolver.cut_cell(atoms,
                                                 vacuum=2.50)

# Initialize GPAW
gs_calc = GPAW(gpts          = gpts,
               experimental={'niter_fixdensity': 2},
               eigensolver   = 'cg',
               nbands        = -1,
               poissonsolver = poissonsolver,
               convergence   = {'energy': energy_eps})
atoms.set_calculator(gs_calc)

# Ground state
energy = atoms.get_potential_energy()

# Test ground state
equal(energy, -0.631881, energy_eps * gs_calc.get_number_of_electrons())

# Test floating point arithmetic errors
equal(gs_calc.hamiltonian.poisson.shift_indices_1, [4, 4, 10], 0)
equal(gs_calc.hamiltonian.poisson.shift_indices_2, [8, 8, 16], 0)

# Save state
gs_calc.write('gs.gpw', 'all')
classical_material = None
gs_calc = None

# Initialize TDDFT and FDTD
kick = [0.0, 0.0, 1.0e-3]
time_step = 10.0
max_time = 100 # 0.1 fs

td_calc = TDDFT('gs.gpw')
td_calc.absorption_kick(kick_strength=kick)
td_calc.hamiltonian.poisson.set_kick(kick)

# Propagate TDDFT and FDTD
td_calc.propagate(time_step,  max_time/time_step/2, 'dm.dat', 'td.gpw')

td_calc2 = TDDFT('td.gpw')
td_calc2.propagate(time_step,  max_time/time_step/2, 'dm.dat', 'td.gpw')

# Test
ref_cl_dipole_moment = [  5.25374117e-14,  5.75811267e-14,  3.08349334e-02]
ref_qm_dipole_moment = [  1.78620337e-11, -1.57782578e-11,  5.21368300e-01]
#print("ref_cl_dipole_moment = %s" % td_calc2.hamiltonian.poisson.get_classical_dipole_moment())
#print("ref_qm_dipole_moment = %s" % td_calc2.hamiltonian.poisson.get_quantum_dipole_moment())

tol = 1e-4
equal(td_calc2.hamiltonian.poisson.get_classical_dipole_moment(), ref_cl_dipole_moment, tol)
equal(td_calc2.hamiltonian.poisson.get_quantum_dipole_moment(), ref_qm_dipole_moment, tol)



from ase import Atoms
from gpaw import GPAW
from gpaw.mixer import FFTMixer
from gpaw.wavefunctions.pw import PW
from gpaw.test import equal

bulk = Atoms('Li', pbc=True,
             cell=[2.6, 2.6, 2.6])
k = 4
bulk.calc = GPAW(mode=PW(200),
                 kpts=(k, k, k),
                 mixer=FFTMixer(),
                 eigensolver='rmm-diis')
e = bulk.get_potential_energy()
equal(e, -1.98481281259, 1.0e-4)
